# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2022 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import socket
import sys
import os
import random
from lib389.paths import Paths
from lib389._constants import INSTALL_LATEST_CONFIG
from lib389.utils import get_default_db_lib, get_default_mdb_max_size, socket_check_bind

MAJOR, MINOR, _, _, _ = sys.version_info

import configparser

format_keys = [
    'prefix',
    'bin_dir',
    'sbin_dir',
    'sysconf_dir',
    'data_dir',
    'local_state_dir',
    'lib_dir',
    'cert_dir',
    'config_dir',
    'inst_dir',
    'backup_dir',
    'db_dir',
    'db_home_dir',
    'db_lib',
    'ldapi',
    'ldif_dir',
    'mdb_max_size',
    'lock_dir',
    'log_dir',
    'run_dir',
    'schema_dir',
    'tmp_dir',
]

ds_paths = Paths()


class Options2(object):
    # This stores the base options in a self._options dict.
    # It provides a number of options for:
    # - dict overlay
    # - parsing the config parser types.

    def get_default():
        # Get option default values if these values are dynamic
        # Concerned options are:
        #   - selinux
        #   - systemd
        #   - port
        #   - secure_port
        if os.geteuid() == 0:
            return { 'selinux': True,
                     'systemd': True,
                     'port': 389,
                     'secure_port': 636 }
        values = { 'selinux': False, 'systemd': False }
        # Try first to select a couple of port like (n*1000+389, n*1000+636)
        for delta in range(1000, 65000, 1000):
            values['port'] = 389+delta
            values['secure_port'] = 636+delta
            if socket_check_bind(values['port']) and socket_check_bind(values['secure_port']):
                return values;
        # Nothing found, so lets use a couple of random values
        for tryid in range(10000):
            values['port'], values['secure_port'] = random.choices(range(1000,65535), k=2)
            if socket_check_bind(values['port']) and socket_check_bind(values['secure_port']):
                return values;
        # Still nothing found.
        # Cannot not raise an exception as this code is also triggered while building 389ds
        # So lets silently choose a couple of ports (probably busy) and get a failure later
        # on (when starting the instance if in dscreate case).
        values['port'], values['secure_port'] = random.choices(range(1000,65535), k=2)
        return values;

    default_values = get_default()

    def get_systemd_default():
        # Cannot use ds_paths.with_systemd in get_default() because dse template
        # is not available in build. So a function is used instead.
        if Options2.default_values['systemd']:
            return ds_paths.with_systemd
        else:
            return False


    def __init__(self, log):
        # 'key' : (default, helptext, valid_func )
        self._options = {}  # this takes the default
        self._type = {}  # Lists the type the item should be.
        self._helptext = {}  # help text for the option, MANDATORY.
        self._advanced = {} # Is the option advanced/developer?
        self._valid_func = {}  # options verification function.
        self._section = None
        self.log = log

    def parse_inf_config(self, config):
        v = None
        for k in self._options.keys():
            try:
                if self._type[k] == int:
                    v = config.getint(self._section, k)
                elif self._type[k] == bool:
                    v = config.getboolean(self._section, k)
                elif self._type[k] == str:
                    v = config.get(self._section, k)
                # How does this handle wrong types?
            except ValueError as e:
                # Invalid type
                err_msg = ('Invalid value in section "%s", "%s" has incorrect type (%s)' % (self._section, k, str(e)))
                raise ValueError(err_msg)
            except configparser.NoOptionError:
                self.log.debug('%s:%s not in inf, using default', self._section, k)
                continue
            self._options[k] = v

    def set(self, option, value):
        self._options[option] = value

    def verify(self):
        pass

    def collect(self):
        return self._options

    def collect_help(self, comment=False, advanced=False):
        helptext = "[%s]\n" % self._section
        for k in sorted(self._options.keys()):
            if self._advanced.get(k, False) is True and advanced is False:
                # Skip this option because we haven't reqed advanced.
                continue
            helptext += "# %s (%s)\n" % (k, self._type[k].__name__)
            helptext += "# Description: %s\n" % (self._helptext[k])
            helptext += "# Default value: %s \n" % (self._options[k])
            helptext += ";%s = %s\n\n" % (k, self._options[k])
        return helptext

#
# Base, example dicts of the general, backend (userRoot) options.
#


class General2Base(Options2):
    def __init__(self, log):
        super(General2Base, self).__init__(log)
        self._section = 'general'

        self._options['config_version'] = 2
        self._type['config_version'] = int
        self._helptext['config_version'] = "Sets the format version of this INF file. To use the INF file with dscreate, you must set this parameter to \"2\"."
        self._advanced['config_version'] = True

        self._options['full_machine_name'] = socket.getfqdn()
        self._type['full_machine_name'] = str
        self._helptext['full_machine_name'] = "Sets the fully qualified hostname (FQDN) of this system. When installing this instance with GSSAPI authentication behind a load balancer, set this parameter to the FQDN of the load balancer and, additionally, set \"strict_host_checking\" to \"false\"."

        self._options['strict_host_checking'] = False
        self._type['strict_host_checking'] = bool
        self._helptext['strict_host_checking'] = "Sets whether the server verifies the forward and reverse record set in the \"full_machine_name\" parameter. When installing this instance with GSSAPI authentication behind a load balancer, set this parameter to \"false\". Container installs imply \"false\"."

        self._options['selinux'] = Options2.default_values['selinux']
        self._type['selinux'] = bool
        self._helptext['selinux'] = "Enables SELinux detection and integration during the installation of this instance. If set to \"True\", dscreate auto-detects whether SELinux is enabled. Set this parameter only to \"False\" in a development environment."
        self._advanced['selinux'] = True

        self._options['systemd'] = Options2.get_systemd_default()
        self._type['systemd'] = bool
        self._helptext['systemd'] = "Enables systemd platform features. If set to \"True\", dscreate auto-detects whether systemd is installed. Set this only to \"False\" in a development environment."
        self._advanced['systemd'] = True

        self._options['start'] = True
        self._type['start'] = bool
        self._helptext['start'] = "Starts the instance after the install completes. If false, the instance is created but not started."

        self._options['defaults'] = INSTALL_LATEST_CONFIG
        self._type['defaults'] = str
        self._helptext['defaults'] = "Directory Server enables administrators to use the default values for cn=config entries from a specific version. If you set this parameter to \"{LATEST}\", which is the default, the instance always uses the default values of the latest version. For example, to configure that the instance uses default values from version 1.3.5, set this parameter to \"001003005\". The format of this value is XXXYYYZZZ, where X is the major version, Y the minor version, and Z the patch level. Note that each part of the value uses 3 digits and must be filled with leading zeros if necessary.".format(LATEST=INSTALL_LATEST_CONFIG)

#
# This module contains the base options and configs for Directory Server
# setup and install. This allows
#


class Slapd2Base(Options2):
    def __init__(self, log):
        super(Slapd2Base, self).__init__(log)
        self._section = 'slapd'

        self._options['instance_name'] = 'localhost'
        self._type['instance_name'] = str
        self._helptext['instance_name'] = "Sets the name of the instance. You can refer to this value in other parameters of this INF file using the \"{instance_name}\" variable. Note that this name cannot be changed after the installation!"

        self._options['user'] = ds_paths.user
        self._type['user'] = str
        self._helptext['user'] = "Sets the user name the ns-slapd process will use after the service started."
        self._advanced['user'] = True

        self._options['group'] = ds_paths.group
        self._type['group'] = str
        self._helptext['group'] = "Sets the group name the ns-slapd process will use after the service started."
        self._advanced['group'] = True

        self._options['root_dn'] = ds_paths.root_dn
        self._type['root_dn'] = str
        self._helptext['root_dn'] = "Sets the Distinquished Name (DN) of the administrator account for this instance. It is recommended that you do not change this value from the default \"cn=Directory Manager\""
        self._advanced['root_dn'] = True

        self._options['root_password'] = 'Directory_Manager_Password'
        self._type['root_password'] = str
        self._helptext['root_password'] = ("Sets the password of the \"cn=Directory Manager\" account (\"root_dn\" parameter)." +
                                           "You can either set this parameter to a plain text password dscreate hashes " +
                                           "during the installation or to a \"{algorithm}hash\" string generated by the " +
                                           "pwdhash utility. The password must be at least 8 characters long.  Note " +
                                           "that setting a plain text password can be a security risk if unprivileged " +
                                           "users can read this INF file!")

        self._options['prefix'] = ds_paths.prefix
        self._type['prefix'] = str
        self._helptext['prefix'] = "Sets the file system prefix for all other directories. You can refer to this value in other fields using the {prefix} variable or the $PREFIX environment variable. Only set this parameter in a development environment."
        self._advanced['prefix'] = True

        self._options['port'] = Options2.default_values['port']
        self._type['port'] = int
        self._helptext['port'] = "Sets the TCP port the instance uses for LDAP connections."

        self._options['secure_port'] = Options2.default_values['secure_port']
        self._type['secure_port'] = int
        self._helptext['secure_port'] = "Sets the TCP port the instance uses for TLS-secured LDAP connections (LDAPS)."

        self._options['self_sign_cert'] = True
        self._type['self_sign_cert'] = bool
        self._helptext['self_sign_cert'] = "Sets whether the setup creates a self-signed certificate and enables TLS encryption during the installation. The certificate is not suitable for production, but it enables administrators to use TLS right after the installation. You can replace the self-signed certificate with a certificate issued by a Certificate Authority. If set to False, you can enable TLS later by importing a CA/Certificate and enabling 'dsconf <instance_name> config replace nsslapd-security=on'"

        self._options['self_sign_cert_valid_months'] = 24
        self._type['self_sign_cert_valid_months'] = int
        self._helptext['self_sign_cert_valid_months'] = "Set the number of months the issued self-signed certificate will be valid."

        # In the future, make bin and sbin /usr/[s]bin, but we may need autotools assistance from Ds
        self._options['bin_dir'] = ds_paths.bin_dir
        self._type['bin_dir'] = str
        self._helptext['bin_dir'] = "Sets the location where the Directory Server binaries are stored. Only set this parameter in a development environment."
        self._advanced['bin_dir'] = True

        self._options['sbin_dir'] = ds_paths.sbin_dir
        self._type['sbin_dir'] = str
        self._helptext['sbin_dir'] = "Sets the location where the Directory Server administration binaries are stored. Only set this parameter in a development environment."
        self._advanced['sbin_dir'] = True

        self._options['sysconf_dir'] = ds_paths.sysconf_dir
        self._type['sysconf_dir'] = str
        self._helptext['sysconf_dir'] = "Sets the location of the system's configuration directory. Only set this parameter in a development environment."
        self._advanced['sysconf_dir'] = True

        self._options['initconfig_dir'] = ds_paths.initconfig_dir
        self._type['initconfig_dir'] = str
        self._helptext['initconfig_dir'] = "Sets the directory of the operating system's rc configuration directory. Only set this parameter in a development environment."
        self._advanced['initconfig_dir'] = True

        # In the future, make bin and sbin /usr/[s]bin, but we may need autotools assistance from Ds
        self._options['data_dir'] = ds_paths.data_dir
        self._type['data_dir'] = str
        self._helptext['data_dir'] = "Sets the location of Directory Server shared static data. Only set this parameter in a development environment."
        self._advanced['data_dir'] = True

        self._options['local_state_dir'] = ds_paths.local_state_dir
        self._type['local_state_dir'] = str
        self._helptext['local_state_dir'] = "Sets the location of Directory Server variable data. Only set this parameter in a development environment."
        self._advanced['local_state_dir'] = True

        self._options['ldapi'] = ds_paths.ldapi
        self._type['ldapi'] = str
        self._helptext['ldapi'] = "Sets the location of socket interface of the Directory Server."

        self._options['lib_dir'] = ds_paths.lib_dir
        self._type['lib_dir'] = str
        self._helptext['lib_dir'] = "Sets the location of Directory Server shared libraries. Only set this parameter in a development environment."
        self._advanced['lib_dir'] = True

        self._options['cert_dir'] = ds_paths.cert_dir
        self._type['cert_dir'] = str
        self._helptext['cert_dir'] = "Sets the directory of the instance's Network Security Services (NSS) database."
        self._advanced['cert_dir'] = True

        self._options['config_dir'] = ds_paths.config_dir
        self._type['config_dir'] = str
        self._helptext['config_dir'] = "Sets the configuration directory of the instance."
        self._advanced['config_dir'] = True

        self._options['inst_dir'] = ds_paths.inst_dir
        self._type['inst_dir'] = str
        self._helptext['inst_dir'] = "Sets the directory of instance-specific scripts."
        self._advanced['inst_dir'] = True

        self._options['backup_dir'] = ds_paths.backup_dir
        self._type['backup_dir'] = str
        self._helptext['backup_dir'] = "Set the backup directory of the instance."
        self._advanced['backup_dir'] = True

        self._options['db_dir'] = ds_paths.db_dir
        self._type['db_dir'] = str
        self._helptext['db_dir'] = "Sets the database directory of the instance."
        self._advanced['db_dir'] = True

        self._options['db_home_dir'] = ds_paths.db_home_dir
        self._type['db_home_dir'] = str
        self._helptext['db_home_dir'] = "Sets the memory-mapped database files location of the instance."
        self._advanced['db_home_dir'] = True

        self._options['db_lib'] = get_default_db_lib()
        self._type['db_lib'] = str
        self._helptext['db_lib'] = "Select the database implementation library (bdb or mdb)."
        self._advanced['db_lib'] = False

        self._options['mdb_max_size'] = get_default_mdb_max_size(ds_paths)
        self._type['mdb_max_size'] = str
        self._helptext['mdb_max_size'] = "Select the lmdb database maximum size."
        self._advanced['mdb_max_size'] = False

        self._options['ldif_dir'] = ds_paths.ldif_dir
        self._type['ldif_dir'] = str
        self._helptext['ldif_dir'] = "Sets the LDIF export and import directory of the instance."
        self._advanced['ldif_dir'] = True

        self._options['lock_dir'] = ds_paths.lock_dir
        self._type['lock_dir'] = str
        self._helptext['lock_dir'] = "Sets the lock directory of the instance."
        self._advanced['lock_dir'] = True

        self._options['log_dir'] = ds_paths.log_dir
        self._type['log_dir'] = str
        self._helptext['log_dir'] = "Sets the log directory of the instance."
        self._advanced['log_dir'] = True

        self._options['run_dir'] = ds_paths.run_dir
        self._type['run_dir'] = str
        self._helptext['run_dir'] = "Sets PID directory of the instance."
        self._advanced['run_dir'] = True

        self._options['schema_dir'] = ds_paths.schema_dir
        self._type['schema_dir'] = str
        self._helptext['schema_dir'] = "Sets schema directory of the instance."
        self._advanced['schema_dir'] = True

        self._options['tmp_dir'] = ds_paths.tmp_dir
        self._type['tmp_dir'] = str
        self._helptext['tmp_dir'] = "Sets the temporary directory of the instance."
        self._advanced['tmp_dir'] = True

    def _format(self, d):
        new_d = {}
        ks = d.keys()
        no_format_keys = set(ks) - set(format_keys)

        for k in no_format_keys:
            new_d[k] = d[k]
        for k in format_keys:
            # Will these be done in correct order?
            if self._type[k] == str:
                new_d[k] = d[k].format(**new_d)
            else:
                new_d[k] = d[k]
        return new_d

    def collect(self):
        # This does the final format and return of options.
        return self._format(self._options)


class Backend2Base(Options2):
    def __init__(self, log, section):
        super(Backend2Base, self).__init__(log)
        self._section = section

        # Suffix settings
        self._options['suffix'] = ''
        self._type['suffix'] = str
        self._helptext['suffix'] = ("Sets the root suffix stored in this database.  If you do not uncomment and set the suffix " +
                                    "attribute the install process will NOT create the backend/suffix.  You can also " +
                                    "create multiple backends/suffixes by duplicating this section.")

        self._options['create_suffix_entry'] = False
        self._type['create_suffix_entry'] = bool
        self._helptext['create_suffix_entry'] = ("Set this parameter to \"True\" to create a generic root node " +
                                                 "entry for the suffix in the database.")

        self._options['sample_entries'] = "no"
        self._type['sample_entries'] = str
        self._helptext['sample_entries'] = ("Set this parameter to 'yes' to add latest version of sample " +
                                            "entries to this database.  Or, use '001003006' to use the " +
                                            "1.3.6 version sample entries.  Use this option, for example, " +
                                            "to create a database for testing purposes.")

        self._options['require_index'] = False
        self._type['require_index'] = bool
        self._helptext['require_index'] = "Set this parameter to \"True\" to refuse unindexed searches in this database."

        # Replication settings
        self._options['enable_replication'] = False
        self._type['enable_replication'] = bool
        self._helptext['enable_replication'] = ("Enable replication for this backend.  By default it will setup the backend as " +
                                                "a supplier, with replica ID 1, and \"cn=replication manager,cn=config\" as the " +
                                                "replication binddn.")

        self._options['replica_role'] = "supplier"
        self._type['replica_role'] = str
        self._helptext['replica_role'] = "Set the replication role.  Choose either 'supplier', 'hub', or 'consumer'"

        self._options['replica_id'] = "1"
        self._type['replica_id'] = int
        self._helptext['replica_id'] = "Set the unique replication identifier for this replica's database (suppliers only)"

        self._options['replica_binddn'] = "cn=replication manager,cn=config"
        self._type['replica_binddn'] = str
        self._helptext['replica_binddn'] = "Set the replication manager DN"

        self._options['replica_bindpw'] = ""
        self._type['replica_bindpw'] = str
        self._helptext['replica_bindpw'] = ("Sets the password of the Replication Manager account (\"replica_binddn\" parameter)." +
                                            "Note that setting a plain text password can be a security risk if unprivileged " +
                                            "users can read this INF file!")

        self._options['replica_bindgroup'] = ""
        self._type['replica_bindgroup'] = str
        self._helptext['replica_bindgroup'] = "Set the replication bind group DN"

        self._options['changelog_max_age'] = "7d"
        self._type['changelog_max_age'] = str
        self._helptext['changelog_max_age'] = ("How long an entry should remain in the replication changelog.  The default is 7 days, or '7d'. (requires that replication is enabled).")

        self._options['changelog_max_entries'] = "-1"
        self._type['changelog_max_entries'] = str
        self._helptext['changelog_max_entries'] = ("The maximum number of entries to keep in the replication changelog.  The default is '-1', which means unlimited. (requires that replication is enabled).")



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019 Red Hat, Inc.
# Copyright (C) 2019 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
import ldap.dn
from ldap.controls import SimplePagedResultsControl
from ldap import filter as ldap_filter
import logging
import json
from functools import partial
from lib389._entry import Entry
from lib389._constants import DIRSRV_STATE_ONLINE
from lib389._mapped_object_lint import DSLint, DSLints
from lib389.utils import (
        ensure_bytes, ensure_str, ensure_int, ensure_list_bytes, ensure_list_str,
        ensure_list_int, display_log_value, display_log_data, is_a_dn, normalizeDN
        )

# This function filter and term generation provided thanks to
# The University of Adelaide. <william@adelaide.edu.au>


def _term_gen(term):
    while True:
        yield term


def _gen(op=None, extra=None):
    filt = ''
    if type(extra) == list:
        for ext in extra:
            filt += ext
    elif type(extra) == str:
        filt += extra
    if filt != '':
        filt = '(%s%s)' % (op, filt)
    return filt


def _gen_and(extra=None):
    return _gen('&', extra)


def _gen_or(extra=None):
    return _gen('|', extra)


def _gen_not(extra=None):
    return _gen('!', extra)


def _gen_filter(attrtypes, values, extra=None):
    filt = ''
    if attrtypes is None:
        raise ValueError("Attempting to filter on type that doesn't support filtering!")
    for attr, value in zip(attrtypes, values):
        if attr is not None and value is not None:
            filt += '(%s=%s)' % (attr, ldap_filter.escape_filter_chars(value))
    if extra is not None:
        filt += '{FILT}'.format(FILT=extra)
    return filt


def _normalise_attrs(attrs_dict):
    result = {}

    for k, v in attrs_dict.items():
        key = k.lower()
        # v is always a list
        values = v
        normalised = []
        for x in values:
            # Handle DN valued attributes
            if is_a_dn(x):
                x = ldap.dn.dn2str(ldap.dn.str2dn(x))
            normalised.append(ensure_str(x))
        result[key] = normalised

    return result


# Define wrappers around the ldap operation to have a clear diagnostic
def _ldap_op_s(inst, f, fname, *args, **kwargs):
    # f.__name__ says 'inner' so the wanted name is provided as argument
    try:
        return f(*args, **kwargs)
    except ldap.LDAPError as e:
        new_desc = f"{fname}({args},{kwargs}) on instance {inst.serverid}"
        if len(e.args) >= 1:
            e.args[0]['ldap_request'] = new_desc
            logging.getLogger().debug(f"args={e.args}")
        raise e

def _add_ext_s(inst, *args, **kwargs):
    return _ldap_op_s(inst, inst.add_ext_s, 'add_ext_s', *args, **kwargs)

def _modify_ext_s(inst, *args, **kwargs):
    return _ldap_op_s(inst, inst.modify_ext_s, 'modify_ext_s', *args, **kwargs)

def _delete_ext_s(inst, *args, **kwargs):
    return _ldap_op_s(inst, inst.delete_ext_s, 'delete_ext_s', *args, **kwargs)

def _search_ext_s(inst, *args, **kwargs):
    return _ldap_op_s(inst, inst.search_ext_s, 'search_ext_s', *args, **kwargs)

def _search_s(inst, *args, **kwargs):
    return _ldap_op_s(inst, inst.search_s, 'search_s', *args, **kwargs)


class DSLogging(object):
    """The benefit of this is automatic name detection, and correct application
    of level and verbosity to the object.

    :param verbose: False by default
    :type verbose: bool
    """

    def __init__(self, verbose=False):
        # Maybe we can think of a way to make this display the instance name or __unicode__?
        self._log = logging.getLogger(type(self).__name__)
        if verbose:
            self._log.setLevel(logging.DEBUG)
        else:
            self._log.setLevel(logging.INFO)


class DSLdapObject(DSLogging, DSLint):
    """A single instance of DSLdapObjects

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    # Outcome results for ensure_state operations
    ENSURE_ADDED = "added"
    ENSURE_UPDATED = "updated"
    ENSURE_UNCHANGED= "unchanged"

    # TODO: Automatically create objects when they are requested to have properties added
    def __init__(self, instance, dn=None):
        self._instance = instance
        super(DSLdapObject, self).__init__(self._instance.verbose)
        # This allows some factor objects to be overriden
        self._dn = None
        if dn is not None:
            self._dn = ensure_str(dn)

        self._protected = True
        # Used in creation
        self._create_objectclasses = []
        self._rdn_attribute = None
        self._must_attributes = None
        # Backend-generated attributes that are local to an instance.
        self._compare_exclude = ['entryid', 'modifytimestamp', 'nsuniqueid', 'parentid',
                                 'numsubordinates', 'hassubordinates', 'tombstonenumsubordinates']
        self._server_controls = None
        self._client_controls = None
        self._object_filter = '(objectClass=*)'
        self._ensure_status = self.ENSURE_UNCHANGED

    def __unicode__(self):
        val = self._dn
        if self._rdn_attribute:
            # What if the rdn is multi value and we don't get the primary .... ARGHHH
            # TODO: https://github.com/389ds/389-ds-base/issues/3329
            val = self.get_rdn_from_dn()
        return ensure_str(val)

    def get_rdn_from_dn(self, dn=None):
        """Extract the RDN value from a DN.

        :param dn: DN to extract RDN from, defaults to this object's DN
        :type dn: str
        :returns: The RDN value
        :rtype: str
        """
        if dn is None:
            dn = self._dn

        rdn_components = ldap.dn.str2dn(dn)[0]
        rdn_value = rdn_components[0][1]
        return rdn_value

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return f'{type(self)}(instance_name="{self._instance.serverid}", dn="{self.__unicode__()}")'

    def _unsafe_raw_entry(self):
        """Get an Entry object

        :returns: Entry object
        """

        return _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=["*"],
                                           serverctrls=self._server_controls, clientctrls=self._client_controls,
                                           escapehatch='i am sure')[0]

    def exists(self):
        """Check if the entry exists

        :returns: True if it exists
        """

        try:
            _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrsonly=1,
                                        serverctrls=self._server_controls, clientctrls=self._client_controls,
                                        escapehatch='i am sure')
        except ldap.NO_SUCH_OBJECT:
            return False

        return True

    def search(self, scope="subtree", filter='objectclass=*'):
        search_scope = ldap.SCOPE_SUBTREE
        if scope == 'base':
            search_scope = ldap.SCOPE_BASE
        elif scope == 'one':
            search_scope = ldap.SCOPE_ONELEVEL
        elif scope == 'subtree':
            search_scope = ldap.SCOPE_SUBTREE
        return _search_ext_s(self._instance,self._dn, search_scope, filter,
                                           serverctrls=self._server_controls,
                                           clientctrls=self._client_controls,
                                           escapehatch='i am sure')

    def display(self, attrlist=['*']):
        """Get an entry but represent it as a string LDIF

        :returns: LDIF formatted string
        """
        e = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=attrlist,
                                        serverctrls=self._server_controls, clientctrls=self._client_controls,
                                        escapehatch='i am sure')
        if len(e) > 0:
            return e[0].__repr__()
        else:
            return ""

    def display_attr(self, attr):
        """Get all values of given attribute - 'attr: value'

        :returns: Formatted string
        """

        out = ""
        for v in self.get_attr_vals_utf8(attr):
            out += "%s: %s\n" % (attr, v)
        return out

    def _jsonify(self, fn, *args, **kwargs):
        # This needs to map all the values to ensure_str
        attrs = fn(use_json=True, *args, **kwargs)
        str_attrs = {}
        for k in attrs:
            str_attrs[ensure_str(k)] = ensure_list_str(attrs[k])

        # ensure all the keys are lowercase
        str_attrs = dict((k.lower(), v) for k, v in list(str_attrs.items()))

        response = json.dumps({"type": "entry", "dn": ensure_str(self._dn), "attrs": str_attrs}, indent=4)

        return response

    def __getattr__(self, name):
        """This enables a bit of magic to allow us to wrap any function ending with
        _json to it's form without json, then transformed. It means your function
        *must* return it's values as a dict of:

        { attr : [val, val, ...], attr : [], ... }
        to be supported.
        """

        if (name.endswith('_json')):
            int_name = name.replace('_json', '')
            pfunc = partial(self._jsonify, getattr(self, int_name))
            return pfunc
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

    # We make this a property so that we can over-ride dynamically if needed
    @property
    def dn(self):
        """Get an object DN

        :returns: DN
        """

        return self._dn

    @property
    def rdn(self):
        """Get an object RDN

        :returns: RDN
        """

        # How can we be sure this returns the primary one?
        # TODO: https://github.com/389ds/389-ds-base/issues/3329
        return ensure_str(self.get_rdn_from_dn())

    def get_basedn(self):
        """Get the suffix this entry belongs to
        """
        from lib389.backend import Backends
        backends = Backends(self._instance).list()
        for backend in backends:
            suffix = backend.get_suffix()
            if self._dn.endswith(suffix):
                return suffix
        return ""

    def present(self, attr, value=None):
        """Assert that some attr, or some attr / value exist on the entry.

        :param attr: an attribute name
        :type attr: str
        :param value: an attribute value
        :type value: str

        :returns: True if attr is present
        """

        if self._instance.state != DIRSRV_STATE_ONLINE:
            raise ValueError("Invalid state. Cannot get presence on instance that is not ONLINE")
        self._log.debug("%s present(%r) %s" % (self._dn, attr, value))

        _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=[attr, ],
                                        serverctrls=self._server_controls, clientctrls=self._client_controls,
                                        escapehatch='i am sure')[0]
        values = self.get_attr_vals_utf8(attr)
        self._log.debug("%s contains %s" % (self._dn, values))

        if value is None:
            # We are just checking if SOMETHING is present ....
            return len(values) > 0

        # Otherwise, we are checking a specific value
        if is_a_dn(value):
            normalized_value = normalizeDN(value)
        else:
            normalized_value = ensure_bytes(value).lower()

        # Normalize each returned value depending on whether it is a DN
        normalized_values = []
        for v in values:
            if is_a_dn(v):
                normalized_values.append(normalizeDN(v))
            else:
                normalized_values.append(ensure_bytes(v.lower()))

        return normalized_value in normalized_values

    def add(self, key, value):
        """Add an attribute with a value

        :param key: an attribute name
        :type key: str
        :param value: an attribute value
        :type value: str
        """

        self.set(key, value, action=ldap.MOD_ADD)

    def add_many(self, *args):
        """Add many key, value pairs in a single operation.
        This is useful for configuration changes that require
        atomic operation, and ease of use.

        An example of usage is add_many((key, value), (key, [value1, value2]))

        No wrapping list is needed for the arguments.

        :param *args: tuples of key,value to add. Value can be a single value
                    or a collection (list, tuple, set) of values.
        :type *args: (str, str) or (str, list/tuple/set)
        """
        mods = []
        for arg in args:
            key, value = arg
            if isinstance(value, (list, tuple, set)):
                value = ensure_list_bytes(list(value))
            else:
                value = [ensure_bytes(value)]

            mods.append((ldap.MOD_ADD, ensure_str(key), value))

        return _modify_ext_s(self._instance,
                            self._dn,
                            mods,
                            serverctrls=self._server_controls,
                            clientctrls=self._client_controls,
                            escapehatch='i am sure')

    def replace(self, key, value):
        """Replace an attribute with a value

        :param key: an attribute name
        :type key: str
        :param value: an attribute value
        :type value: str
        """
        self.set(key, value, action=ldap.MOD_REPLACE)

    def replace_many(self, *args):
        """Replace many key, value pairs in a single operation.
        This is useful for configuration changes that require
        atomic operation, and ease of use.

        An example of usage is replace_many((key, value), (key, [value1, value2]))

        No wrapping list is needed for the arguments.

        :param *args: tuples of key,value to replace. Value can be a single value
                    or a collection (list, tuple, set) of values.
        :type *args: (str, str) or (str, list/tuple/set)
        """
        mods = []
        for arg in args:
            key, value = arg
            if isinstance(value, (list, tuple, set)):
                value = ensure_list_bytes(list(value))
            else:
                value = [ensure_bytes(value)]

            mods.append((ldap.MOD_REPLACE, ensure_str(key), value))

        return _modify_ext_s(self._instance,
                            self._dn,
                            mods,
                            serverctrls=self._server_controls,
                            clientctrls=self._client_controls,
                            escapehatch='i am sure')

    # This needs to work on key + val, and key
    def remove(self, key, value):
        """Remove a value defined by key

        :param key: an attribute name
        :type key: str
        :param value: an attribute value
        :type value: str
        """

        # Do a mod_delete on the value.
        self.set(key, value, action=ldap.MOD_DELETE)

    def remove_all(self, key):
        """Remove all values defined by key (if possible).

        If an attribute is multi-valued AND required all values except one will
        be deleted.

        :param key: an attribute name
        :type key: str
        """

        try:
            self.set(key, None, action=ldap.MOD_DELETE)
        except ldap.NO_SUCH_ATTRIBUTE:
            pass

    def ensure_present(self, attr, value):
        """Ensure that an attribute and value are present in a state,
        or add it.

        :param key: an attribute name
        :type key: str
        :param value: an attribute value
        :type value: str
        """
        if not self.present(attr, value):
            self.add(attr, value)

    def ensure_removed(self, attr, value):
        """Ensure that a attribute and value has been removed and not present
        or remove it.

        :param key: an attribute name
        :type key: str
        :param value: an attribute value
        :type value: str
        """
        if self.present(attr, value):
            self.remove(attr, value)

    @property
    def ensure_status(self):
        return self._ensure_status

    def ensure_attr_state(self, state):
        """
        Given a dict of attr-values, ensure they are in the same state on the entry. This is
        a stateful assertion, generally used by things like PATCH in a REST api.

        The format is:
            {
                'attr_1': ['value', 'value'],
                'attr_2': [],
            }

        If a value is present in the list, but not in the entry it is ADDED.
        If a value is NOT present in the list, and is on the entry, it is REMOVED.
        If a value is an empty list [], the attr is REMOVED from the entry.
        If an attr is not named in the dictionary, it is not altered.

        This function is atomic - all changes are applied or none are. There are no
        partial updates.

        This function is idempotent - submitting the same request twice will cause no
        action to be taken as we are ensuring a state, not listing actions to take.

        :param state: The entry ava state
        :type state: dict
        """
        self._log.debug('ensure_state')
        # Get all our entry/attrs in a single batch
        entry_state = self.get_attrs_vals_utf8(state.keys())

        # Check what is present/is not present to work out what has to change.
        modlist = []
        for (attr, values) in state.items():
            value_set = set(values)
            entry_set = set(entry_state.get(attr, []))

            # Set difference, is "all items in s but not t".
            value_add = value_set - entry_set
            value_rem = entry_set - value_set

            for value in value_add:
                modlist.append((ldap.MOD_ADD, attr, value))
            for value in value_rem:
                modlist.append((ldap.MOD_DELETE, attr, value))

        self._log.debug('Applying modlist: %s' % modlist)
        # Apply it!
        if len(modlist) > 0:
            self.apply_mods(modlist)

    # maybe this could be renamed?
    def set(self, key, value, action=ldap.MOD_REPLACE):
        """Perform a specified action on a key with value

        :param key: an attribute name
        :type key: str
        :param value: an attribute value
        :type value: str
        :param action: - ldap.MOD_REPLACE - by default
                        - ldap.MOD_ADD
                        - ldap.MOD_DELETE
        :type action: int

        :returns: result of modify_s operation
        :raises: ValueError - if instance is not online
        """

        if action == ldap.MOD_ADD:
            action_txt = "ADD"
        elif action == ldap.MOD_REPLACE:
            action_txt = "REPLACE"
        elif action == ldap.MOD_DELETE:
            action_txt = "DELETE"
        else:
            # This should never happen (bug!)
            action_txt = "UNKNOWN"

        if value is None or len(value) < 512:
            self._log.debug("%s set %s: (%r, %r)" % (self._dn, action_txt, key, display_log_value(key, value)))
        else:
            self._log.debug("%s set %s: (%r, value too large)" % (self._dn, action_txt, key))
        if self._instance.state != DIRSRV_STATE_ONLINE:
            raise ValueError("Invalid state. Cannot set properties on instance that is not ONLINE.")

        if isinstance(value, list):
            # value = map(lambda x: ensure_bytes(x), value)
            value = ensure_list_bytes(value)
        elif value is not None:
            value = [ensure_bytes(value)]

        return _modify_ext_s(self._instance, self._dn, [(action, key, value)],
                             serverctrls=self._server_controls,
                             clientctrls=self._client_controls,
                             escapehatch='i am sure')

    def apply_mods(self, mods):
        """Perform modification operation using several mods at once

        :param mods: [(action, key, value),] or [(ldap.MOD_DELETE, key),]
        :type mods: list of tuples
        :raises: ValueError - if a provided mod op is invalid
        """

        mod_list = []
        for mod in mods:
            if len(mod) < 2:
                # Error
                raise ValueError('Not enough arguments in the mod op')
            elif len(mod) == 2:  # no action
                # This hack exists because the original lib389 Entry type
                # does odd things.
                action, key = mod
                if action != ldap.MOD_DELETE:
                    raise ValueError('Only MOD_DELETE takes two arguments %s' % mod)
                value = None
                # Just add the raw mod, because we don't have a value
                mod_list.append((action, key, value))
            elif len(mod) == 3:
                action, key, value = mod
                if action != ldap.MOD_REPLACE and \
                   action != ldap.MOD_ADD and \
                   action != ldap.MOD_DELETE:
                    raise ValueError('Invalid mod action(%s)' % str(action))
                if isinstance(value, list):
                    value = ensure_list_bytes(value)
                else:
                    value = [ensure_bytes(value)]
                mod_list.append((action, key, value))
            else:
                # Error too many items
                raise ValueError('Too many arguments in the mod op')
        return _modify_ext_s(self._instance,self._dn, mod_list, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')

    def _unsafe_compare_attribute(self, other):
        """Compare two attributes from two objects. This is currently marked unsafe as it's
        not complete yet.

        The idea is to use the native ldap compare operation, rather than simple comparison of
        values due to schema awareness. LDAP doesn't normalise values, so two objects with:

        cn: value
        cn: VaLuE

        This will fail to compare in python, but would succeed in LDAP beacuse CN is case
        insensitive.

        To allow schema aware checking, we need to call ldap compare extop here.
        """
        pass

    @classmethod
    def compare(cls, obj1, obj2):
        """Compare if two RDN objects have same attributes and values.

        This comparison is a loose comparison, not a strict one i.e. "this object *is* this other object"
        It will just check if the attributes are same.
        Instance-local operational attributes like 'nsUniqueId', 'entryid',
        and 'parentid' are not checked intentionally because we want to
        compare arbitrary objects, i.e they may have different internal
        database identity but same attributes.

        Example::

            cn=user1,ou=a
            cn=user1,ou=b

        Comparision of these two objects should result in same, even though their 'nsUniqueId' attribute differs.

        :param obj1: An entry to check
        :type obj1: lib389._mapped_object.DSLdapObject
        :param obj2: An entry to check
        :type obj2: lib389._mapped_object.DSLdapObject
        :returns: True if objects have same attributes else returns False
        :raises: ValueError - if obj1 or obj2 don't inherit DSLdapObject
        """

        # ensuring both the objects are RDN objects
        if not issubclass(type(obj1), DSLdapObject) or not issubclass(type(obj2), DSLdapObject):
            raise ValueError("Invalid arguments: Expecting object types that inherits 'DSLdapObject' class")
        # check if RDN of objects is same
        if obj1.rdn != obj2.rdn:
            obj1._log.debug("%s != %s" % (obj1.rdn, obj2.rdn))
            return False
        obj1_attrs = obj1.get_compare_attrs()
        obj2_attrs = obj2.get_compare_attrs()
        # Bail fast if the keys don't match
        if set(obj1_attrs.keys()) != set(obj2_attrs.keys()):
            obj1._log.debug("%s != %s" % (obj1_attrs.keys(), obj2_attrs.keys()))
            return False
        # Check the values of each key
        # using obj1_attrs.keys() because obj1_attrs.iterkleys() is not supported in python3
        for key in obj1_attrs.keys():
            if set(obj1_attrs[key]) != set(obj2_attrs[key]):
                obj1._log.debug("  v-- %s != %s" % (key, key))
                obj1._log.debug("%s != %s" % (obj1_attrs[key], obj2_attrs[key]))
                return False
        return True

    def get_compare_attrs(self, use_json=False):
        """Get a dictionary having attributes to be compared
        i.e. excluding self._compare_exclude
        """

        self._log.debug("%s get_compare_attrs" % (self._dn))

        all_attrs_dict = self.get_all_attrs()
        all_attrs_lower = {}
        for k in all_attrs_dict:
            all_attrs_lower[k.lower()] = all_attrs_dict[k]

        # removing _compate_exclude attrs from all attrs
        cx = [x.lower() for x in self._compare_exclude]
        compare_attrs = set(all_attrs_lower.keys()) - set(cx)

        compare_attrs_dict = {attr.lower(): all_attrs_lower[attr] for attr in compare_attrs}

        return compare_attrs_dict

    def get_all_attrs(self, use_json=False):
        """Get a dictionary having all the attributes of the entry

        :returns: Dict with real attributes and operational attributes
        """

        self._log.debug("%s get_all_attrs" % (self._dn))
        if self._instance.state != DIRSRV_STATE_ONLINE:
            raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
        else:
            # retrieving real(*) and operational attributes(+)
            attrs_entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                      attrlist=["*", "+"], serverctrls=self._server_controls,
                                                      clientctrls=self._client_controls, escapehatch='i am sure')
            if len(attrs_entry) > 0:
                # getting dict from 'entry' object
                attrs_dict = attrs_entry[0].data
                # Should we normalise the attr names here to lower()?
                # This could have unforseen consequences ...
                return attrs_dict
            else:
                return {}

    def get_all_attrs_utf8(self, use_json=False):
        """Get a dictionary having all the attributes of the entry

        :returns: Dict with real attributes and operational attributes
        """

        self._log.debug("%s get_all_attrs" % (self._dn))
        if self._instance.state != DIRSRV_STATE_ONLINE:
            raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
        else:
            # retrieving real(*) and operational attributes(+)
            attrs_entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                      attrlist=["*", "+"], serverctrls=self._server_controls,
                                                      clientctrls=self._client_controls, escapehatch='i am sure')
            if len(attrs_entry) > 0:
                # getting dict from 'entry' object
                r = {}
                for (k, vo) in attrs_entry[0].data.items():
                    r[k] = ensure_list_str(vo)
                return r
            else:
                return {}

    def get_attrs_vals(self, keys, use_json=False):
        self._log.debug("%s get_attrs_vals(%r)" % (self._dn, keys))
        if self._instance.state != DIRSRV_STATE_ONLINE:
            raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
        else:
            entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                attrlist=keys, serverctrls=self._server_controls,
                                                clientctrls=self._client_controls, escapehatch='i am sure')
            if len(entry) > 0:
                return entry[0].getValuesSet(keys)
            else:
                return []

    def get_attrs_vals_utf8(self, keys, use_json=False):
        self._log.debug("%s get_attrs_vals_utf8(%r)" % (self._dn, keys))
        if self._instance.state != DIRSRV_STATE_ONLINE:
            raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
        entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter, attrlist=keys,
                                            serverctrls=self._server_controls, clientctrls=self._client_controls,
                                            escapehatch='i am sure')
        if len(entry) > 0:
            vset = entry[0].getValuesSet(keys)
            r = {}
            for (k, vo) in vset.items():
                r[k] = ensure_list_str(vo)
            return r
        else:
            return {}

    def get_attr_vals(self, key, use_json=False):
        self._log.debug("%s get_attr_vals(%r)" % (self._dn, key))
        # We might need to add a state check for NONE dn.
        if self._instance.state != DIRSRV_STATE_ONLINE:
            raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
            # In the future, I plan to add a mode where if local == true, we
            # can use get on dse.ldif to get values offline.
        else:
            # It would be good to prevent the entry code intercepting this ....
            # We have to do this in this method, because else we ignore the scope base.
            entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                attrlist=[key], serverctrls=self._server_controls,
                                                clientctrls=self._client_controls, escapehatch='i am sure')
            if len(entry) > 0:
                vals = entry[0].getValues(key)
                if use_json:
                    result = {key: []}
                    for val in vals:
                        result[key].append(val)
                    return result
                else:
                    return vals
            else:
                if use_json:
                    return {}
                else:
                    return []

    def get_attr_val(self, key, use_json=False):
        self._log.debug("%s getVal(%r)" % (self._dn, key))
        # We might need to add a state check for NONE dn.
        if self._instance.state != DIRSRV_STATE_ONLINE:
            raise ValueError("Invalid state. Cannot get properties on instance that is not ONLINE")
            # In the future, I plan to add a mode where if local == true, we
            # can use get on dse.ldif to get values offline.
        else:
            entry = _search_ext_s(self._instance,self._dn, ldap.SCOPE_BASE, self._object_filter,
                                                attrlist=[key], serverctrls=self._server_controls,
                                                clientctrls=self._client_controls, escapehatch='i am sure')
            if len(entry) > 0:
                return entry[0].getValue(key)
            else:
                return ""

    def get_attr_val_bytes(self, key, use_json=False):
        """Get a single attribute value from the entry in bytes type

        :param key: An attribute name
        :type key: str
        :returns: A single bytes value
        :raises: ValueError - if instance is offline
        """

        return ensure_bytes(self.get_attr_val(key))

    def get_attr_vals_bytes(self, key, use_json=False):
        """Get attribute values from the entry in bytes type

        :param key: An attribute name
        :type key: str
        :returns: A list of bytes values
        :raises: ValueError - if instance is offline
        """

        return ensure_list_bytes(self.get_attr_vals(key))

    def get_attr_val_utf8(self, key, use_json=False):
        """Get a single attribute value from the entry in utf8 type

        :param key: An attribute name
        :type key: str
        :returns: A single UTF8 value
        :raises: ValueError - if instance is offline
        """

        return ensure_str(self.get_attr_val(key))

    def get_attr_val_utf8_l(self, key, use_json=False):
        """Get a single attribute value from the entry in utf8 type

        :param key: An attribute name
        :type key: str
        :returns: A single lowered UTF8 value
        :raises: ValueError - if instance is offline
        """

        x = self.get_attr_val(key)
        if x is not None:
            return ensure_str(x).lower()
        else:
            return None

    def get_attr_vals_utf8(self, key, use_json=False):
        """Get attribute values from the entry in utf8 type

        :param key: An attribute name
        :type key: str
        :returns: A list of UTF8 values
        :raises: ValueError - if instance is offline
        """

        return ensure_list_str(self.get_attr_vals(key))

    def get_attr_vals_utf8_l(self, key, use_json=False):
        """Get attribute values from the entry in utf8 type and lowercase

        :param key: An attribute name
        :type key: str
        :returns: A list of lowered UTF8 values
        :raises: ValueError - if instance is offline
        """

        return [x.lower() for x in ensure_list_str(self.get_attr_vals(key))]

    def get_attr_val_int(self, key, use_json=False):
        """Get a single attribute value from the entry in int type

        :param key: An attribute name
        :type key: str
        :returns: A single int value
        :raises: ValueError - if instance is offline
        """

        return ensure_int(self.get_attr_val(key))

    def get_attr_vals_int(self, key, use_json=False):
        """Get attribute values from the entry in int type

        :param key: An attribute name
        :type key: str
        :returns: A list of int values
        :raises: ValueError - if instance is offline
        """

        return ensure_list_int(self.get_attr_vals(key))

    # Duplicate, but with many values. IE a dict api.
    # This
    def add_values(self, values):
        pass

    def replace_values(self, values):
        pass

    def set_values(self, values, action=ldap.MOD_REPLACE):
        pass

    # Modifies the DN of an entry to the new fqdn provided
    def rename(self, new_rdn, newsuperior=None, deloldrdn=True):
        """Renames the object within the tree.

        If you provide a newsuperior, this will move the object in the tree.
        If you only provide a new_rdn, it stays in the same branch, but just
        changes the rdn.

        Note, if you use newsuperior, you may move this object outside of the
        scope of the related DSLdapObjects manager, which may cause it not to
        appear in .get() requests.

        :param new_rdn: RDN of the new entry
        :type new_rdn: str
        :param newsuperior: New parent DN
        :type newsuperior: str
        """
        # When we are finished with this, we need to update our DN
        # To do this, we probably need to search the new rdn as a filter,
        # and the superior as the base (if it changed)
        if self._protected:
            return

        self._instance.rename_s(self._dn, new_rdn, newsuperior,
                                serverctrls=self._server_controls, clientctrls=self._client_controls,
                                delold=deloldrdn, escapehatch='i am sure')
        if newsuperior is not None:
            # Well, the new DN should be rdn + newsuperior.
            self._dn = '%s,%s' % (new_rdn, newsuperior)
        else:
            old_dn_parts = ldap.explode_dn(self._dn)
            # Replace the rdn
            old_dn_parts[0] = new_rdn
            self._dn = ",".join(old_dn_parts)
        assert self.exists()

        # assert we actually got the change right ....

    def delete(self, recursive=False):
        """Deletes the object defined by self._dn.
        This can be changed with the self._protected flag!
        """

        self._log.debug("%s delete" % (self._dn))
        if not self._protected:
            # Is there a way to mark this as offline and kill it
            if recursive:
                filterstr = "(|(objectclass=*)(objectclass=ldapsubentry))"
                ents = _search_s(self._instance, self._dn, ldap.SCOPE_SUBTREE, filterstr, escapehatch='i am sure')
                for ent in sorted(ents, key=lambda e: len(e.dn), reverse=True):
                    _delete_ext_s(self._instance, ent.dn, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
            else:
                _delete_ext_s(self._instance, self._dn, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')

    def _validate(self, rdn, properties, basedn):
        """Used to validate a create request.
        This way, it can be over-ridden without affecting
        the create types.

        It also checks that all the values in _must_attribute exist
        in some form in the dictionary.

        It has the useful trick of returning the dn, so subtypes
        can use extra properties to create the dn's here for this.
        """

        if properties is None:
            raise ldap.UNWILLING_TO_PERFORM('Invalid request to create. Properties cannot be None')
        if type(properties) != dict:
            raise ldap.UNWILLING_TO_PERFORM("properties must be a dictionary")

        # I think this needs to be made case insensitive
        # How will this work with the dictionary?
        if self._must_attributes is not None:
            for attr in self._must_attributes:
                if properties.get(attr, None) is None:
                    # Put RDN to properties
                    if attr == self._rdn_attribute and rdn is not None:
                        properties[self._rdn_attribute] = self.get_rdn_from_dn(rdn)
                    else:
                        raise ldap.UNWILLING_TO_PERFORM('Attribute %s must not be None' % attr)

        # Make sure the naming attribute is present
        if properties.get(self._rdn_attribute, None) is None and rdn is None:
            raise ldap.UNWILLING_TO_PERFORM('Attribute %s must not be None or rdn provided' % self._rdn_attribute)

        # Great! Now, lets fix up our types
        for k, v in properties.items():
            if isinstance(v, list):
                # Great!
                pass
            else:
                # Turn it into a list instead.
                properties[k] = [v, ]

        # If we were created with a dn= in the object init, we set tdn now, and skip
        # any possible dn derivation steps that follow.
        tdn = self._dn

        # However, if no DN was provided, we attempt to derive the DN from the relevant
        # properties of the object. The idea being that by defining only the attributes
        # of the object, we can just solve a possible rdn instead of asking for the same
        # data twice.
        if tdn is None:
            if basedn is None:
                raise ldap.UNWILLING_TO_PERFORM('Invalid request to create. basedn cannot be None')

            # Were we given a relative component? Yes? Go ahead!
            if rdn is not None:
                tdn = ensure_str('%s,%s' % (rdn, basedn))
            elif properties.get(self._rdn_attribute, None) is not None:
                # Favour the value in the properties dictionary
                v = properties.get(self._rdn_attribute)
                rdn = ensure_str(v[0])

                erdn = ensure_str(ldap.dn.escape_dn_chars(rdn))
                self._log.debug("Using first property %s: %s as rdn" % (self._rdn_attribute, erdn))
                # Now we compare. If we changed this value, we have to put it back to make the properties complete.
                if erdn != rdn:
                    properties[self._rdn_attribute].append(erdn)

                tdn = ensure_str('%s=%s,%s' % (self._rdn_attribute, erdn, basedn))

        # We may need to map over the data in the properties dict to satisfy python-ldap
        str_props = {}
        for k, v in properties.items():
            str_props[k] = ensure_list_bytes(v)
        #
        # Do we need to do extra dn validation here?
        return (tdn, str_props)

    def _create(self, rdn=None, properties=None, basedn=None, ensure=False, strict=False):
        """Internal implementation of create. This is used by ensure
        and create, to prevent code duplication. You should *never* call
        this method directly.
        """
        assert(len(self._create_objectclasses) > 0)
        basedn = ensure_str(basedn)
        self._log.debug('Checking "%s" under %s : %s' % (rdn, basedn, display_log_data(properties)))
        # Add the objectClasses to the properties
        (dn, valid_props) = self._validate(rdn, properties, basedn)
        # Check if the entry exists or not? .add_s is going to error anyway ...
        self._log.debug('Validated dn {}'.format(dn))

        exists = False

        if ensure:
            # If we are running in stateful ensure mode, we need to check if the object exists, and
            # we can see the state that it is in.
            try:
                _search_ext_s(self._instance,dn, ldap.SCOPE_BASE, self._object_filter, attrsonly=1, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
                exists = True
            except ldap.NO_SUCH_OBJECT:
                pass

        if exists and ensure:
            # update properties
            self._log.debug('Exists %s' % dn)
            self._dn = dn

            # Non strict mode, replace all attributes
            if not strict:
                # Now use replace_many to setup our values
                mods = []
                for k, v in list(valid_props.items()):
                    mods.append((ldap.MOD_REPLACE, k, v))
                _modify_ext_s(self._instance,self._dn, mods, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')

                self._ensure_status = self.ENSURE_UPDATED
                return self
            # Strict mode, compare and update only the differences
            else:
                # Get existing attributes
                current_attrs = _normalise_attrs(self.get_all_attrs_utf8())

                # Get requested attributes
                requested_attrs = _normalise_attrs(valid_props)

                # Remove operational attributes
                for attr in self._compare_exclude:
                    current_attrs.pop(attr, None)

                # Compare current with requested
                matches = True
                mods = []
                for k, v in requested_attrs.items():
                    current_val = current_attrs.get(k, [])
                    if set(current_val) != set(v):
                        matches = False
                        mods.append((ldap.MOD_REPLACE, k, ensure_list_bytes(v)))

                # Its a match, nothing to do
                if matches:
                    self._ensure_status = self.ENSURE_UNCHANGED
                    return self

                if mods:
                    _modify_ext_s(self._instance, self._dn, mods,
                                serverctrls=self._server_controls,
                                clientctrls=self._client_controls,
                                escapehatch='i am sure')
                    self._ensure_status = self.ENSURE_UPDATED

                return self

        elif not exists:
            # This case is reached in two cases. One is we are in ensure mode, and we KNOW the entry
            # doesn't exist.
            # The alternate, is that we are in a non-stateful create, so we "just create" and see
            # what happens. I believe the technical term is "yolo create".
            self._log.debug('Creating %s' % dn)
            e = Entry(dn)
            e.update({'objectclass': ensure_list_bytes(self._create_objectclasses)})
            e.update(valid_props)
            # We rely on exceptions here to indicate failure to the parent.
            _add_ext_s(self._instance, e, serverctrls=self._server_controls, clientctrls=self._client_controls, escapehatch='i am sure')
            self._log.debug('Created entry %s : %s' % (dn, display_log_data(e.data)))
            # If it worked, we need to fix our instance dn for the object's self reference. Because
            # we may not have a self reference yet (just created), it may have changed (someone
            # set dn, but validate altered it).
            self._dn = dn
            self._ensure_status = self.ENSURE_ADDED
            return self
        else:
            # This case can't be reached now that we only check existance on ensure.
            # However, it's good to keep it for "complete" behaviour, exhausting all states.
            # We could highlight bugs ;)
            raise AssertionError("Impossible State Reached in _create")

    def create(self, rdn=None, properties=None, basedn=None):
        """Add a new entry

        :param rdn: RDN of the new entry
        :type rdn: str
        :param properties: Attributes for the new entry
        :type properties: dict
        :param basedn: Base DN of the new entry
        :type rdn: str

        :returns: DSLdapObject of the created entry
        """
        return self._create(rdn, properties, basedn, ensure=False)

    def ensure_state(self, rdn=None, properties=None, basedn=None, strict=False):
        """Ensure an entry exists with the following state, created
        if necessary.

        :param rdn: RDN of the new entry
        :type rdn: str
        :param properties: Attributes for the new entry
        :type properties: dict
        :param basedn: Base DN of the new entry
        :type rdn: str

        :returns: DSLdapObject of the created entry
        """
        return self._create(rdn, properties, basedn, ensure=True, strict=strict)


# A challenge of this, is how do we manage indexes? They have two naming attributes....

class DSLdapObjects(DSLogging, DSLints):
    """The object represents the next idea: "Everything is an instance of something
    that exists in this way", i.e. we unite LDAP entries by some
    set of parameters with the object.

    :param instance: An instance
    :type instance: lib389.DirSrv
    """

    def __init__(self, instance, basedn=""):
        self._childobject = DSLdapObject
        self._instance = instance
        super(DSLdapObjects, self).__init__(self._instance.verbose)
        self._objectclasses = []
        self._filterattrs = []
        self._list_attrlist = ['dn']
        # Copy this from the child if we need.
        self._basedn = basedn
        self._scope = ldap.SCOPE_SUBTREE
        self._server_controls = None
        self._client_controls = None

    def _get_objectclass_filter(self):
        return _gen_and(
            _gen_filter(_term_gen('objectclass'), self._objectclasses)
        )

    def _get_selector_filter(self, selector):
        return _gen_and([
            self._get_objectclass_filter(),
            _gen_or(
                # This will yield all combinations of selector to filterattrs.
                # This won't work with multiple values in selector (yet)
                _gen_filter(self._filterattrs, _term_gen(selector))
            ),
        ])

    def _entry_to_instance(self, dn=None, entry=None):
        # Normally this won't be used. But for say the plugin type where we
        # have "many" possible child types, this allows us to overload
        # and select / return the right one through ALL our get/list/create
        # functions with very little work on the behalf of the overloader
        return self._childobject(instance=self._instance, dn=dn)

    def list(self, paged_search=None, paged_critical=True, full_dn=False):
        """Get a list of children entries (DSLdapObject, Replica, etc.) using
        a base DN and objectClasses of our object (DSLdapObjects, Replicas,
        etc.)

        :param paged_search: None for no paged search, or an int of page size
                             to use.
        :param paged_critical: pages search is critical
        :param full_dn: Return a list of DN's instead of objects
        :returns: A list of children entries
        """

        # Filter based on the objectclasses and the basedn
        insts = None
        # This will yield an & filter for objectClass with as many terms as
        # needed.
        filterstr = self._get_objectclass_filter()
        self._log.debug('list filter = %s' % filterstr)

        if type(paged_search) == int:
            self._log.debug('listing with paged search -> %d', paged_search)
            # If paged_search ->
            results = []
            pages = 0
            pctrls = []
            req_pr_ctrl = SimplePagedResultsControl(paged_critical,
                                                    size=paged_search,
                                                    cookie='')
            if self._server_controls is not None:
                controls = [req_pr_ctrl] + self._server_controls
            else:
                controls = [req_pr_ctrl]
            while True:
                msgid = self._instance.search_ext(
                        base=self._basedn,
                        scope=self._scope,
                        filterstr=filterstr,
                        attrlist=self._list_attrlist,
                        serverctrls=controls,
                        clientctrls=self._client_controls,
                        escapehatch='i am sure'
                    )
                self._log.info('Getting page %d' % (pages,))
                rtype, rdata, rmsgid, rctrls = self._instance.result3(msgid,
                                                                      escapehatch='i am sure')
                results.extend(rdata)
                pages += 1
                self._log.debug("%s" % rctrls)
                pctrls = [c for c in rctrls if c.controlType == SimplePagedResultsControl.controlType]
                if pctrls and pctrls[0].cookie:
                    req_pr_ctrl.cookie = pctrls[0].cookie
                    if self._server_controls is not None:
                        controls = [req_pr_ctrl] + self._server_controls
                    else:
                        controls = [req_pr_ctrl]
                else:
                    break
                # End while
            # Result3 doesn't map through Entry, so we have to do it manually.
            results = [Entry(r) for r in results]
            if full_dn:
                insts = [r.dn for r in results]
            else:
                insts = [self._entry_to_instance(dn=r.dn, entry=r) for r in results]
            # End paged search
        else:
            # If not paged
            try:
                results = _search_ext_s(self._instance,
                                        base=self._basedn,
                                        scope=self._scope,
                                        filterstr=filterstr,
                                        attrlist=self._list_attrlist,
                                        serverctrls=self._server_controls,
                                        clientctrls=self._client_controls,
                                        escapehatch='i am sure')
                if full_dn:
                    insts = [r.dn for r in results]
                else:
                    insts = [self._entry_to_instance(dn=r.dn, entry=r) for r in results]
            except ldap.NO_SUCH_OBJECT:
                # There are no objects to select from, se we return an empty
                # array
                insts = []
        return insts

    def exists(self, selector=[], dn=None):
        """Check if a child entry exists

        :returns: True if it exists
        """
        results = []
        try:
            if dn is not None:
                results = self._get_dn(dn)
            else:
                results = self._get_selector(selector)
        except:
            return False

        if len(results) == 1:
            return True
        else:
            return False

    def get(self, selector=[], dn=None, json=False):
        """Get a child entry (DSLdapObject, Replica, etc.) with dn or selector
        using a base DN and objectClasses of our object (DSLdapObjects, Replicas, etc.)

        Note that * is not a valid selector, you should use "list()" instead.

        :param dn: DN of wanted entry
        :type dn: str
        :param selector: An additional filter to search for, i.e. 'backend_name'. The attributes selected are based on object type, ie user will search for uid and cn.
        :type dn: str

        :returns: A child entry
        """

        results = []
        if dn is not None:
            criteria = dn
            search_filter = self._get_objectclass_filter()
            results = self._get_dn(dn)
        else:
            criteria = selector
            search_filter = self._get_selector_filter(selector)
            results = self._get_selector(selector)

        if len(results) == 0:
            raise ldap.NO_SUCH_OBJECT(f"No object exists given the filter criteria: {criteria} {search_filter}")
        if len(results) > 1:
            entry_dn = [e.dn for e in results]
            entry_dns_pretty = '\n    '.join(entry_dn)
            raise ldap.UNWILLING_TO_PERFORM(f"Too many objects matched selection criteria: {criteria} {search_filter}"
                                            f" - Please use 'get-by-dn' to specify which entry to get:\n    {entry_dns_pretty}")
        if json:
            return self._entry_to_instance(results[0].dn, results[0]).get_all_attrs_json()
        else:
            return self._entry_to_instance(results[0].dn, results[0])

    def _get_dn(self, dn):
        # This will yield and & filter for objectClass with as many terms as needed.
        filterstr = self._get_objectclass_filter()
        self._log.debug('_gen_dn filter = %s' % filterstr)
        self._log.debug('_gen_dn dn = %s' % dn)
        return _search_ext_s(self._instance,
            base=dn,
            scope=ldap.SCOPE_BASE,
            filterstr=filterstr,
            attrlist=self._list_attrlist,
            serverctrls=self._server_controls, clientctrls=self._client_controls,
            escapehatch='i am sure'
        )

    def _get_selector(self, selector):
        # Filter based on the objectclasses and the basedn
        # Based on the selector, we should filter on that too.
        # This will yield and & filter for objectClass with as many terms as needed.
        filterstr = self._get_selector_filter(selector)
        self._log.debug('_gen_selector filter = %s' % filterstr)
        return _search_ext_s(self._instance,
            base=self._basedn,
            scope=self._scope,
            filterstr=filterstr,
            attrlist=self._list_attrlist,
            serverctrls=self._server_controls, clientctrls=self._client_controls,
            escapehatch='i am sure'
        )

    def _validate(self, rdn, properties):
        """Validate the factory part of the creation"""

        if properties is None:
            raise ldap.UNWILLING_TO_PERFORM('Invalid request to create. Properties cannot be None')
        if type(properties) != dict:
            raise ldap.UNWILLING_TO_PERFORM("properties must be a dictionary")

        return (rdn, properties)

    def create(self, rdn=None, properties=None):
        """Create an object under base DN of our entry

        :param rdn: RDN of the new entry
        :type rdn: str
        :param properties: Attributes for the new entry
        :type properties: dict

        :returns: DSLdapObject of the created entry
        """

        # Should we inject the rdn to properties?
        # This may not work in all cases, especially when we consider plugins.
        #
        co = self._entry_to_instance(dn=None, entry=None)
        # Make the rdn naming attr available
        self._rdn_attribute = co._rdn_attribute
        (rdn, properties) = self._validate(rdn, properties)
        # Now actually commit the creation req
        return co.create(rdn, properties, self._basedn)

    def ensure_state(self, rdn=None, properties=None, strict=False):
        """Create an object under base DN of our entry, or
        assert it exists and update it's properties.

        :param rdn: RDN of the new entry
        :type rdn: str
        :param properties: Attributes for the new entry
        :type properties: dict

        :returns: DSLdapObject of the created entry
        """

        # Should we inject the rdn to properties?
        # This may not work in all cases, especially when we consider plugins.
        #
        co = self._entry_to_instance(dn=None, entry=None)
        # Make the rdn naming attr avaliable
        self._rdn_attribute = co._rdn_attribute
        (rdn, properties) = self._validate(rdn, properties)
        # Now actually commit the creation req
        return co.ensure_state(rdn, properties, self._basedn, strict=strict)

    def filter(self, search, attrlist=None, scope=None, strict=False):
        # This will yield and & filter for objectClass with as many terms as needed.
        if search:
            search_filter = _gen_and([self._get_objectclass_filter(), search])
        else:
            search_filter = self._get_objectclass_filter()
        if scope is None:
            scope = self._scope
        if attrlist:
            self._list_attrlist = attrlist
        self._log.debug(f'list filter = {search_filter} with scope {scope} and attribute list {attrlist}')
        try:
            results = _search_ext_s(self._instance,
                base=self._basedn,
                scope=scope,
                filterstr=search_filter,
                attrlist=self._list_attrlist,
                serverctrls=self._server_controls, clientctrls=self._client_controls
            )
            # def __init__(self, instance, dn=None):
            insts = [self._entry_to_instance(dn=r.dn, entry=r) for r in results]
        except ldap.NO_SUCH_OBJECT:
            # There are no objects to select from, se we return an empty array
            if strict:
                raise ldap.NO_SUCH_OBJECT
            insts = []
        return insts

class CompositeDSLdapObject(DSLdapObject):
    """A virtual view as a single object that merges two entry attributes.
       This class is not supposed to be called directly but through subclasses.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Global dn
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(CompositeDSLdapObject, self).__init__(instance, dn)
        self._entries =[]
        self._map = {}
        self.log = None

    def add_component(self, entry, attrs):
        """Add a new entry as element of the composite entry.
           An attribute may belong to different entries,
           if that is the case:
                the attribute is written in all entries
                the attribute is read from last entry containing it.

        :param entry: An entry object
        :type entry: DSLdapObject
        :param attrs: Attributes from the entry that are in the composite entry.
        :type attrs: (str,...)
        """

        idx = len(self._entries)
        self._entries.append(entry)
        for attr in attrs:
            nattr = attr.lower()
            if nattr in self._map:
                self._map[nattr].append(idx)
            else:
                self._map[nattr] = [idx,]

    def _find_idx(self, attr, must_have=True):
        """Find the entry index."""
        nattr = ensure_str(attr).lower()
        try:
            return self._map[nattr]
        except KeyError:
            if must_have:
                raise ValueError(f'No mapping for attribute {attr} in composite object')
            return []

    def _unsafe_raw_entry(self):
        raise NotImplementedError("Composite objects have no raw view.")

    def exists(self):
        for entry in self._entries:
            if not entry.exists():
                return False
        return True

    def search(self, scope="subtree", filter='objectclass=*'):
        raise NotImplementedError("Cannot use this method on composite objects.")

    def get_basedn(self):
        raise NotImplementedError("Cannot use this method on composite objects.")

    def present(self, attr, value=None):
        for entry in self._entries:
            if not entry.present():
                return False
        return True

    def _spread_set(self, dataset, set2attr_fn):
        """Separate a set of data to a set per entry."""
        res = []
        for entry in self._entries:
            res.append([])
        for data in dataset:
            attr = set2attr_fn(data)
            for idx in self._find_idx(attr):
                res[idx].append(data)
        return res

    def _spread_dict(self, datadict):
        """Separate a dict of data where key is the attribure to a dict per entry."""
        res = {}
        for entry in self._entries:
            res.append({})
        for attr in datadict.keys():
            for idx in self._find_idx(attr):
                res[idx][attr] = datadict[attr]
        return res

    def replace_many(self, *args):
        dataset = self._spread_set(*args, lambda x: x[0])
        for idx in range(0, self._entries):
            if len(dataset[idx]) > 0:
                self._entries[idx].replace_many(*dataset[idx])

    def ensure_attr_state(self, state):
        # Hard to implement because some requires attributes may be common to several entries
        raise NotImplementedError("Cannot use this method on composite objects.")

    def set(self, key, value, action=ldap.MOD_REPLACE):
        v = { ldap.MOD_REPLACE : "REPLACE",
              ldap.MOD_ADD : "ADD",
              ldap.MOD_DELETE: "DEL" }
        for idx in self._find_idx(key):
            self._entries[idx].set(key, value, action)

    def apply_mods(self, mods):
        dataset = self._spread_set(*args, lambda x: x[1])
        for idx in range(0, self._entries):
            if len(dataset[idx]) > 0:
                self._entries[idx].apply_mods(*dataset[idx])

    def get_all_attrs(self, use_json=False):
        res = {}
        idx = 0
        for entry in self._entries:
            attrs = entry.get_all_attrs(use_json)
            for attr in attrs.keys():
                if idx in self._find_idx(attr, False):
                    res[attr] = attrs[attr]
            idx += 1
        return res

    def get_all_attrs_utf8(self, use_json=False):
        res = {}
        idx = 0
        for entry in self._entries:
            attrs = entry.get_all_attrs_utf8(use_json)
            for attr in attrs.keys():
                if idx in self._find_idx(attr, False):
                    res[attr] = attrs[attr]
            idx += 1
        return res

    def get_attrs_vals(self, keys, use_json=False):
        res = {}
        dataset = self._spread_set(*args, lambda x: x)
        for idx in range(0, self._entries):
            if len(dataset[idx]) > 0:
                res.update(self._entries[idx].apply_mods(*dataset[idx], use_json))

    def get_attrs_vals_utf8(self, keys, use_json=False):
        res = {}
        dataset = self._spread_set(*args, lambda x: x)
        for idx in range(0, self._entries):
            if len(dataset[idx]) > 0:
                res.update(self._entries[idx].apply_mods(*dataset[idx], use_json))

    def get_attr_vals(self, key, use_json=False):
        idx = self._find_idx(key)[-1]
        return self._entries[idx].get_attr_vals(key, use_json)

    def get_attr_vals_utf8(self, key, use_json=False):
        idx = self._find_idx(key)[-1]
        return self._entries[idx].get_attr_vals_utf8(key, use_json)

    def get_attr_val(self, key, use_json=False):
        idx = self._find_idx(key)[-1]
        return self._entries[idx].get_attr_val(key, use_json)

    def get_attr_val_utf8(self, key, use_json=False):
        idx = self._find_idx(key)[-1]
        return self._entries[idx].get_attr_val_utf8(key, use_json)

    def add_values(self, values):
        raise DeprecationWarning("Not implemented any more.")

    def replace_values(self, values):
        raise DeprecationWarning("Not implemented any more.")

    def set_values(self, values, action=ldap.MOD_REPLACE):
        raise DeprecationWarning("Not implemented any more.")

    def rename(self, new_rdn, newsuperior=None, deloldrdn=True):
        raise NotImplementedError("Cannot use this method on composite objects.")

    def delete(self, recursive=False):
        raise NotImplementedError("Cannot use this method on composite objects.")

    def _validate(self, rdn, properties, basedn):
        raise NotImplementedError("Cannot use this method on composite objects.")

    def _create(self, rdn=None, properties=None, basedn=None, ensure=False):
        raise NotImplementedError("Cannot use this method on composite objects.")

    def create(self, rdn=None, properties=None, basedn=None):
        raise NotImplementedError("Cannot use this method on composite objects.")

    def ensure_state(self, rdn=None, properties=None, basedn=None):
        raise NotImplementedError("Cannot use this method on composite objects.")

    def display(self, attrlist=['*']):
        """Get an entry but represent it as a string LDIF
        :returns: LDIF formatted string
        """
        eattrs = {}
        idx = 0
        for entry in self._entries:
            e = _search_ext_s(entry._instance,entry._dn, ldap.SCOPE_BASE, entry._object_filter, attrlist=attrlist,
                                        serverctrls=entry._server_controls, clientctrls=entry._client_controls,
                                        escapehatch='i am sure')[0]
            for attr, vals in e.iterAttrs():
                if idx in self._find_idx(attr, False):
                    eattrs[attr] = vals
            idx += 1
        e = Entry((self.dn, eattrs))
        return e.__repr__()



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389._mapped_object import DSLdapObjects
# Account derives DSLdapObject - it gives us the lock / unlock functions.
from lib389.idm.account import Account
from lib389.utils import ds_is_older
from lib389.cli_base.dsrc import dsrc_to_ldap
from lib389._constants import DSRC_HOME

MUST_ATTRIBUTES = [
    'uid',
    'cn',
    'sn',
    'uidNumber',
    'gidNumber',
    'homeDirectory',
]
RDN = 'uid'
DEFAULT_BASEDN_RDN = 'ou=People'

TEST_USER_PROPERTIES = {
    'uid': 'testuser',
    'cn' : 'testuser',
    'sn' : 'user',
    'uidNumber' : '1000',
    'gidNumber' : '2000',
    'homeDirectory' : '/home/testuser'
}


# Modern userAccounts

class nsUserAccount(Account):
    _must_attributes = [
        'uid',
        'cn',
        'displayName',
        'uidNumber',
        'gidNumber',
        'homeDirectory',
    ]

    """A single instance of an nsPerson, capable of posix login, certificate
    authentication, sshkey distribution, and more.

    This is the modern and correct userAccount type to choose for DS 1.4.0 and above.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """
    def __init__(self, instance, dn=None):
        if ds_is_older('1.4.0', instance=instance):
            raise Exception("Not supported")
        super(nsUserAccount, self).__init__(instance, dn)
        self._rdn_attribute = RDN
        self._must_attributes = nsUserAccount._must_attributes
        # Can I generate these from schema?
        self._create_objectclasses = [
            'top',
            'nsPerson',
            'nsAccount',
            'nsOrgPerson',
            'posixAccount',
        ]
        user_compare_exclude = [
            'nsUniqueId',
            'modifyTimestamp',
            'createTimestamp',
            'entrydn'
        ]
        self._compare_exclude = self._compare_exclude + user_compare_exclude
        self._protected = False


class nsUserAccounts(DSLdapObjects):
    """DSLdapObjects that represents all nsUserAccount entries in suffix.
    By default it uses 'ou=People' as rdn.

    This is the modern and correct userAccount type to choose for DS 1.4.0 and above.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    :param rdn: The DN that will be combined with basedn
    :type rdn: str
    """

    def __init__(self, instance, basedn, rdn='ou=people'):
        super(nsUserAccounts, self).__init__(instance)
        self._objectclasses = [
            'nsPerson',
            'nsAccount',
            'nsOrgPerson',
            'posixAccount',
        ]
        self._filterattrs = [RDN, 'displayName', 'cn']
        self._childobject = nsUserAccount

        dsrc_inst = dsrc_to_ldap(DSRC_HOME, instance.serverid, self._log)
        if dsrc_inst is not None and 'people_rdn' in dsrc_inst and dsrc_inst['people_rdn'] is not None:
            rdn = dsrc_inst['people_rdn']

        if rdn is None:
            self._basedn = basedn
        else:
            self._basedn = f'{rdn},{basedn}'

    def create_test_user(self, uid=1000, gid=2000):
        """Create a test user with uid=test_user_UID rdn

        :param uid: User id
        :type uid: int
        :param gid: Group id
        :type gid: int

        :returns: DSLdapObject of the created entry
        """

        rdn_value = f"test_user_{uid}"
        rdn = f"uid={rdn_value}"
        properties = {
            'uid': rdn_value,
            'cn': rdn_value,
            'displayName': rdn_value,
            'uidNumber': str(uid),
            'gidNumber': str(gid),
            'homeDirectory': f'/home/{rdn_value}',
        }
        return super(nsUserAccounts, self).create(rdn, properties)


# Traditional style userAccounts.

class UserAccount(Account):
    """A single instance of User Account entry

    This is the classic "user account" style of cn + sn. You should consider
    nsUserAccount instead.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(UserAccount, self).__init__(instance, dn)
        self._rdn_attribute = RDN
        # Can I generate these from schema?
        self._must_attributes = MUST_ATTRIBUTES
        self._create_objectclasses = [
            'top',
            'account',
            'posixaccount',
            'inetOrgPerson',
            'organizationalPerson',
        ]
        if ds_is_older('1.3.7', instance=instance):
            self._create_objectclasses.append('inetUser')
        else:
            self._create_objectclasses.append('nsMemberOf')
        if not ds_is_older('1.4.0', instance=instance):
            self._create_objectclasses.append('nsAccount')
        user_compare_exclude = [
            'nsUniqueId',
            'modifyTimestamp',
            'createTimestamp',
            'entrydn'
        ]
        self._compare_exclude = self._compare_exclude + user_compare_exclude
        self._protected = False

    def _validate(self, rdn, properties, basedn):
        if 'ntUserDomainId' in properties and 'ntUser' not in self._create_objectclasses:
            self._create_objectclasses.append('ntUser')

        return super(UserAccount, self)._validate(rdn, properties, basedn)


class UserAccounts(DSLdapObjects):
    """DSLdapObjects that represents all User Account entries in suffix.
    By default it uses 'ou=People' as rdn.

    This is the classic "user account" style of cn + sn. You should consider
    nsUserAccounts instead.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    :param rdn: The DN that will be combined with basedn
    :type rdn: str
    """

    def __init__(self, instance, basedn, rdn=DEFAULT_BASEDN_RDN):
        super(UserAccounts, self).__init__(instance)
        self._objectclasses = [
            'account',
            'posixaccount',
            'inetOrgPerson',
            'organizationalPerson',
        ]
        self._filterattrs = [RDN]
        self._childobject = UserAccount

        dsrc_inst = dsrc_to_ldap(DSRC_HOME, instance.serverid, self._log)
        if dsrc_inst is not None and 'people_rdn' in dsrc_inst and dsrc_inst['people_rdn'] is not None:
            rdn = dsrc_inst['people_rdn']

        if rdn is None:
            self._basedn = basedn
        else:
            self._basedn = f'{rdn},{basedn}'

    def create_test_user(self, uid=1000, gid=2000):
        """Create a test user with uid=test_user_UID rdn

        :param uid: User id
        :type uid: int
        :param gid: Group id
        :type gid: int

        :returns: DSLdapObject of the created entry
        """

        rdn_value = f"test_user_{uid}"
        rdn = f"uid={rdn_value}"
        properties = {
            'uid': rdn_value,
            'cn': rdn_value,
            'sn': rdn_value,
            'uidNumber': str(uid),
            'gidNumber': str(gid),
            'homeDirectory': f'/home/{rdn_value}'
        }
        return super(UserAccounts, self).create(rdn, properties)


# The following classes are used to manage user accounts created by the UI

class TraditionalUserAccount(Account):
    """A single instance of a TraditionalUser Account entry

    This is the "Traditional" user account created by the UI

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    _must_attributes = ['cn', 'sn']

    def __init__(self, instance, dn=None):
        super(TraditionalUserAccount, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        # Can I generate these from schema?
        self._must_attributes = ['cn', 'sn']
        self._create_objectclasses = [
            'top',
            'person',
            'inetOrgPerson',
            'organizationalPerson',
        ]
        user_compare_exclude = [
            'nsUniqueId',
            'modifyTimestamp',
            'createTimestamp',
            'entrydn'
        ]
        self._compare_exclude = self._compare_exclude + user_compare_exclude
        self._protected = False

    def _validate(self, rdn, properties, basedn):
        if 'ntUserDomainId' in properties and 'ntUser' not in self._create_objectclasses:
            self._create_objectclasses.append('ntUser')

        return super(TraditionalUserAccount, self)._validate(rdn, properties, basedn)


class TraditionalUserAccounts(DSLdapObjects):
    """DSLdapObjects that represents all TraditionalUser Account entries in
    suffix. By default it uses 'ou=People' as rdn.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    :param rdn: The DN that will be combined with basedn
    :type rdn: str
    """

    def __init__(self, instance, basedn, rdn=DEFAULT_BASEDN_RDN):
        super(TraditionalUserAccounts, self).__init__(instance)
        self._objectclasses = [
            'person',
            'inetOrgPerson',
            'organizationalPerson',
        ]
        self._filterattrs = [RDN, 'cn']
        self._childobject = TraditionalUserAccount

        dsrc_inst = dsrc_to_ldap(DSRC_HOME, instance.serverid, self._log)
        if dsrc_inst is not None and 'people_rdn' in dsrc_inst and dsrc_inst['people_rdn'] is not None:
            rdn = dsrc_inst['people_rdn']

        if rdn is None:
            self._basedn = basedn
        else:
            self._basedn = f'{rdn},{basedn}'

    def create_test_user(self):
        """Create a test user with uid=test_user rdn

        :returns: DSLdapObject of the created entry
        """

        rdn_value = "test_user"
        rdn = f"cn={rdn_value}"
        properties = {
            'uid': rdn_value,
            'cn': rdn_value,
            'sn': rdn_value,
        }
        return super(TraditionalUserAccounts, self).create(rdn, properties)

class BasicUserAccount(Account):
    """A single instance of a Basic User Account entry

    This is the "Basic" user account created by the UI

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    _must_attributes = ['uid', 'cn', 'displayName']

    def __init__(self, instance, dn=None):
        super(BasicUserAccount, self).__init__(instance, dn)
        self._rdn_attribute = RDN
        # Can I generate these from schema?
        self._must_attributes = ['uid', 'cn', 'displayName']
        self._create_objectclasses = [
            'top',
            'nsPerson',
            'nsAccount',
            'nsOrgPerson'
        ]
        user_compare_exclude = [
            'nsUniqueId',
            'modifyTimestamp',
            'createTimestamp',
            'entrydn'
        ]
        self._compare_exclude = self._compare_exclude + user_compare_exclude
        self._protected = False

    def _validate(self, rdn, properties, basedn):
        if 'ntUserDomainId' in properties and 'ntUser' not in self._create_objectclasses:
            self._create_objectclasses.append('ntUser')

        return super(BasicUserAccount, self)._validate(rdn, properties, basedn)


class BasicUserAccounts(DSLdapObjects):
    """DSLdapObjects that represents all BasicUser Account entries in
    suffix. By default it uses 'ou=People' as rdn.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    :param rdn: The DN that will be combined with basedn
    :type rdn: str
    """

    def __init__(self, instance, basedn, rdn=DEFAULT_BASEDN_RDN):
        super(BasicUserAccounts, self).__init__(instance)
        self._objectclasses = [
            'nsPerson',
            'nsAccount',
            'nsOrgPerson'
        ]
        self._filterattrs = [RDN]
        self._childobject = BasicUserAccount

        dsrc_inst = dsrc_to_ldap(DSRC_HOME, instance.serverid, self._log)
        if dsrc_inst is not None and 'people_rdn' in dsrc_inst and dsrc_inst['people_rdn'] is not None:
            rdn = dsrc_inst['people_rdn']

        if rdn is None:
            self._basedn = basedn
        else:
            self._basedn = f'{rdn},{basedn}'

    def create_test_user(self):
        """Create a test user with uid=test_user rdn

        :returns: DSLdapObject of the created entry
        """

        rdn_value = "test_user"
        rdn = f"uid={rdn_value}"
        properties = {
            'uid': rdn_value,
            'cn': rdn_value,
            'sn': rdn_value,
        }
        return super(BasicUserAccounts, self).create(rdn, properties)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019, William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389._mapped_object import DSLdapObject, DSLdapObjects
from lib389.utils import ensure_str

class UnsafeExtensibleObject(DSLdapObject):
    """A single instance of an extensible object. Extensible object by it's
    nature is unsafe, eliminating rules around attribute checking. It may
    cause unsafe or other unknown behaviour if not handled correctly.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(UnsafeExtensibleObject, self).__init__(instance, dn)
        self._rdn_attribute = "cn"
        # Can I generate these from schema?
        self._must_attributes = []
        self._create_objectclasses = [
            'top',
            'extensibleObject',
        ]
        self._protected = False

class UnsafeExtensibleObjects(DSLdapObjects):
    """DSLdapObjects that represents all extensible objects. Extensible Objects
    are unsafe in their nature, disabling many checks around schema and attribute
    handling. You should really really REALLY not use this unless you have specific
    needs for testing.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Base DN for all group entries below
    :type basedn: str
    """

    def __init__(self, instance, basedn):
        super(UnsafeExtensibleObjects, self).__init__(instance)
        self._objectclasses = [
            'extensibleObject',
        ]
        self._filterattrs = ["cn"]
        self._childobject = UnsafeExtensibleObject
        self._basedn = ensure_str(basedn)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import pytest
from lib389.utils import *

#
# socket related functions
#
import socket


def test_normalizeDN():
    test = [
        (r'dc=example,dc=com', r'dc=example,dc=com'),
        (r'dc=example, dc=com', r'dc=example,dc=com'),
        (r'cn="dc=example,dc=com",cn=config',
         'cn=dc\\=example\\,dc\\=com,cn=config'),
    ]
    for k, v in test:
        r = normalizeDN(k)
        assert r == v, "Mismatch %r vs %r" % (r, v)


def test_escapeDNValue():
    test = [(r'"dc=example, dc=com"', r'\"dc\=example\,\ dc\=com\"')]
    for k, v in test:
        r = escapeDNValue(k)
        assert r == v, "Mismatch %r vs %r" % (r, v)


def test_escapeDNFiltValue():
    test = [(r'"dc=example, dc=com"',
             '\\22dc\\3dexample\\2c\\20dc\\3dcom\\22')]
    for k, v in test:
        r = escapeDNFiltValue(k)
        assert r == v, "Mismatch %r vs %r" % (r, v)


def test_isLocalHost():
    test = [
        ('localhost', True),
        ('localhost.localdomain', True),
        (socket.gethostname(), True),
        ('www.google.it', False)]
    for k, v in test:
        r = isLocalHost(k)
        assert r == v, "Mismatch %r vs %r on %r" % (r, v, k)


def test_update_newhost_with_fqdn():
    # This test doesn't account for platform where localhost.localdomain isn't
    # the name ....
    test = [
        ({'hostname': 'localhost'}, (['localhost.localdomain', 'localhost'], True)),
        ({'hostname': 'remote'}, (['remote'], False)),
    ]
    for k, v in test:
        old = k.copy()
        expected_host, expected_r = v
        r = update_newhost_with_fqdn(k)
        assert expected_r == r, "Mismatch %r vs %r for %r" % (
            r, expected_r, old)
        assert k['hostname'] in expected_host, "Mismatch %r vs %r for %r" % (
            k['hostname'], expected_host, old)


def test_formatInfData():
    ret = formatInfData({
        'hostname': 'localhost.localdomain',
        'user-id': 'dirsrv',
        'group-id': 'dirsrv',
        'ldap-port': 12345,
        'root-dn': 'cn=directory manager',
        'root-pw': 'password',
        'server-id': 'dirsrv',
        'suffix': 'o=base1',
        'strict_hostname_checking': True,
    })
    log.info("content: %r" % ret)


def test_formatInfData_withadmin():
    instance_params = {
        'hostname': 'localhost.localdomain',
        'user-id': 'dirsrv',
        'group-id': 'dirsrv',
        'ldap-port': 12346,
        'root-dn': 'cn=directory manager',
        'root-pw': 'password',
        'server-id': 'dirsrv',
        'suffix': 'o=base1',
        'strict_hostname_checking': True,
        }
    admin_params = {
        'have_admin': True,
        'create_admin': True,
        'admin_domain': 'example.com',
        'cfgdshost': 'localhost',
        'cfgdsport': 12346,
        'cfgdsuser': 'admin',
        'cfgdspwd': 'admin'}
    instance_params.update(admin_params)
    ret = formatInfData(instance_params)
    log.info("content: %r" % ret)


def test_formatInfData_withconfigserver():
    instance_params = {
        'hostname': 'localhost.localdomain',
        'user-id': 'dirsrv',
        'group-id': 'dirsrv',
        'ldap-port': 12346,
        'root-dn': 'cn=directory manager',
        'root-pw': 'password',
        'server-id': 'dirsrv',
        'suffix': 'o=base1',
        'strict_hostname_checking': True,
        }
    admin_params = {
        'have_admin': True,
        'cfgdshost': 'localhost',
        'cfgdsport': 12346,
        'cfgdsuser': 'admin',
        'cfgdspwd': 'admin',
        'admin_domain': 'example.com'}
    instance_params.update(admin_params)
    ret = formatInfData(instance_params)
    log.info("content: %r" % ret)


@pytest.mark.parametrize('data', [
    ({'userpaSSwoRd': '1234', 'nsslaPd-rootpw': '5678', 'regularAttr': 'originalvalue'},
     {'userpaSSwoRd': '********', 'nsslaPd-rootpw': '********', 'regularAttr': 'originalvalue'}),
    ({'userpassword': ['1', '2', '3'], 'nsslapd-rootpw': ['x']},
     {'userpassword': ['********', '********', '********'], 'nsslapd-rootpw': ['********']})
])
def test_get_log_data(data):
    before, after = data
    assert display_log_data(before) == after


@pytest.mark.parametrize('ds_ver, cmp_ver', [
    ('1.3.1', '1.3.2'),
    ('1.3.1', '1.3.10'),
    ('1.3.2', '1.3.10'),
    ('1.3.9', ('1.3.10', '1.4.2.0')),
    ('1.4.0.1', ('1.3.9', '1.4.1.0', '1.4.2.1')),
    ('1.4.1', '1.4.2.0-20191115gitbadc0ffee' ),
])
def test_ds_is_older_versions(ds_ver, cmp_ver):
    if isinstance(cmp_ver, tuple):
        assert ds_is_related('older', ds_ver, *cmp_ver)
    else:
        assert ds_is_related('older', ds_ver, cmp_ver)

@pytest.mark.parametrize('ds_ver, cmp_ver', [
    ('1.3.2', '1.3.1'),
    ('1.3.10', '1.3.1'),
    ('1.3.10', '1.3.2'),
    ('1.3.10', ('1.3.9', '1.4.2.0')),
    ('1.4.2.1', ('1.3.9', '1.4.0.1', '1.4.2.0')),
    ('1.4.2.0-20191115gitbadc0ffee', '1.4.1' ),
])
def test_ds_is_newer_versions(ds_ver, cmp_ver):
    if isinstance(cmp_ver, tuple):
        assert ds_is_related('newer', ds_ver, *cmp_ver)
    else:
        assert ds_is_related('newer', ds_ver, cmp_ver)


@pytest.mark.parametrize('input, result', [
    (b'', ''),
    (b'\x00', '\\00'),
    (b'\x01\x00', '\\01\\00'),
    (b'01', '\\30\\31'),
    (b'101', '\\31\\30\\31'),
    (b'101x1', '\\31\\30\\31\\78\\31'),
    (b'0\x82\x05s0\x82\x03[\xa0\x03\x02\x01\x02', '\\30\\82\\05\\73\\30\\82\\03\\5b\\a0\\03\\02\\01\\02'),
])
def test_search_filter_escape_bytes(input, result):
    assert search_filter_escape_bytes(input) == result


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -v %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389 import DirSrv, InvalidArgumentError
from lib389.properties import *


class Suffix(object):
    proxied_methods = 'search_s getEntry'.split()

    def __init__(self, conn):
        """@param conn - a DirSrv instance"""
        self.conn = conn
        self.log = conn.log

    def __getattr__(self, name):
        if name in Suffix.proxied_methods:
            return DirSrv.__getattr__(self.conn, name)

    def list(self):
        '''
            Returns the list of suffixes DN for which it exists a mapping tree
            entry

            @param None

            @return list of suffix DN

            @raise None
        '''
        suffixes = []
        ents = self.conn.mappingtree.list()
        for ent in ents:
            vals = self.conn.mappingtree.toSuffix(entry=ent)
            suffixes.append(vals[0])
        return suffixes

    def toBackend(self, suffix=None):
        '''
            Returns the backend entry that stores the provided suffix

            @param suffix - suffix DN of the backend

            @return backend - LDAP entry of the backend

            @return ValueError - if suffix is not provided

        '''
        if not suffix:
            raise ValueError("suffix is mandatory")

        return self.conn.backend.list(suffix=suffix)

    def getParent(self, suffix=None):
        '''
            Returns the DN of a suffix that is the parent of the provided
            'suffix'.
            If 'suffix' has no parent, it returns None

            @param suffix - suffix DN of the backend

            @return parent suffix DN

            @return ValueError - if suffix is not provided
                    InvalidArgumentError - if suffix is not implemented on the
                                           server

        '''
        if not suffix:
            raise ValueError("suffix is mandatory")

        ents = self.conn.mappingtree.list(suffix=suffix)
        if len(ents) == 0:
            raise InvalidArgumentError(
                "suffix %s is not implemented on that server" % suffix)

        mapping_tree = ents[0]
        if mapping_tree.hasValue(MT_PROPNAME_TO_ATTRNAME[MT_PARENT_SUFFIX]):
            return mapping_tree.getValue(
                MT_PROPNAME_TO_ATTRNAME[MT_PARENT_SUFFIX])
        else:
            return None

    def setProperties(self, suffix):
        '''
            Supported properties:

        '''

        if not suffix:
            raise ValueError("suffix is mandatory")



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import logging
import pytest

from lib389._constants import *
from lib389.dseldif import DSEldif
from lib389.tests.topologies import topology_st as topo

DEBUGGING = os.getenv('DEBUGGING', False)

if DEBUGGING:
    logging.getLogger(__name__).setLevel(logging.DEBUG)
else:
    logging.getLogger(__name__).setLevel(logging.INFO)

log = logging.getLogger(__name__)


@pytest.mark.parametrize("entry_dn", (DN_CONFIG,
                                      DN_CONFIG_LDBM))
def test_get_singlevalue(topo, entry_dn):
    """Check that we can get an attribute value under different suffixes"""

    dse_ldif = DSEldif(topo.standalone)

    log.info("Get 'cn' attr from {}".format(entry_dn))
    attr_values = dse_ldif.get(entry_dn, "cn")
    assert attr_values == ["config"]

    log.info("Get 'nonexistent' attr from {}".format(entry_dn))
    attr_values = dse_ldif.get(entry_dn, "nonexistent")
    assert not attr_values


def test_get_multivalue(topo):
    """Check that we can get attribute values"""

    dse_ldif = DSEldif(topo.standalone)

    log.info("Get objectClass from {}".format(DN_CONFIG))
    attr_values = dse_ldif.get(DN_CONFIG, "objectClass")
    assert len(attr_values) == 3
    assert "top" in attr_values
    assert "extensibleObject" in attr_values
    assert "nsslapdConfig" in attr_values


@pytest.mark.parametrize("fake_attr_value", ("fake value",
                                             "fakevalue"))
def test_add(topo, fake_attr_value):
    """Check that we can add an attribute to a given suffix"""

    dse_ldif = DSEldif(topo.standalone)
    fake_attr = "fakeAttr"

    log.info("Add {} to {}".format(fake_attr, DN_CONFIG))
    dse_ldif.add(DN_CONFIG, fake_attr, fake_attr_value)
    attr_values = dse_ldif.get(DN_CONFIG, fake_attr)
    assert attr_values == [fake_attr_value]

    log.info("Clean up")
    dse_ldif.delete(DN_CONFIG, fake_attr)
    assert not dse_ldif.get(DN_CONFIG, fake_attr)


def test_replace(topo):
    """Check that we can replace an attribute to a given suffix"""

    dse_ldif = DSEldif(topo.standalone)
    port_attr = "nsslapd-port"
    port_value = "390"

    log.info("Get default value of {}".format(port_attr))
    default_value = dse_ldif.get(DN_CONFIG, port_attr)[0]

    log.info("Replace {} with {}".format(port_attr, port_value))
    dse_ldif.replace(DN_CONFIG, port_attr, port_value)
    attr_values = dse_ldif.get(DN_CONFIG, port_attr)
    assert attr_values == [port_value]

    log.info("Restore default value")
    dse_ldif.replace(DN_CONFIG, port_attr, default_value)


def test_delete_singlevalue(topo):
    """Check that we can delete an attribute from a given suffix"""

    dse_ldif = DSEldif(topo.standalone)
    fake_attr = "fakeAttr"
    fake_attr_values = ["fake1", "fake2", "fake3"]

    log.info("Add multivalued {} to {}".format(fake_attr, DN_CONFIG))
    for value in fake_attr_values:
        dse_ldif.add(DN_CONFIG, fake_attr, value)

    log.info("Delete {}".format(fake_attr_values[0]))
    dse_ldif.delete(DN_CONFIG, fake_attr, fake_attr_values[0])
    attr_values = dse_ldif.get(DN_CONFIG, fake_attr)
    assert len(attr_values) == 2
    assert fake_attr_values[0] not in attr_values
    assert fake_attr_values[1] in attr_values
    assert fake_attr_values[2] in attr_values

    log.info("Clean up")
    dse_ldif.delete(DN_CONFIG, fake_attr)
    assert not dse_ldif.get(DN_CONFIG, fake_attr)


def test_delete_multivalue(topo):
    """Check that we can delete attributes from a given suffix"""

    dse_ldif = DSEldif(topo.standalone)
    fake_attr = "fakeAttr"
    fake_attr_values = ["fake1", "fake2", "fake3"]

    log.info("Add multivalued {} to {}".format(fake_attr, DN_CONFIG))
    for value in fake_attr_values:
        dse_ldif.add(DN_CONFIG, fake_attr, value)

    log.info("Delete all values of {}".format(fake_attr))
    dse_ldif.delete(DN_CONFIG, fake_attr)
    assert not dse_ldif.get(DN_CONFIG, fake_attr)




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import datetime
import os
import ldap
import logging
import re
import tempfile
from typing import Optional
from lib389._mapped_object import DSLdapObjects, DSLdapObject
from lib389.utils import cert_is_ca, pem_to_der, is_pem_cert, ensure_str
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import pkcs12

log = logging.getLogger(__name__)

DYCATTR_CN    = "cn"
DYNCERT_SUFFIX = "cn=dynamiccertificates"

DYCATTR_PREFIX      = "dsdynamiccertificate"
DYCATTR_CERTDER     = DYCATTR_PREFIX + "der"
DYCATTR_PKEYDER     = DYCATTR_PREFIX + "privatekeyder"
DYCATTR_SUBJECT     = DYCATTR_PREFIX + "subject"
DYCATTR_ISSUER      = DYCATTR_PREFIX + "issuer"
DYCATTR_TRUST       = DYCATTR_PREFIX + "trustflags"
DYCATTR_NOTAFTER    = DYCATTR_PREFIX + "notafter"
DYCATTR_FORCE	    = DYCATTR_PREFIX + "Force"
DYCATTR_ISCA        = DYCATTR_PREFIX + "IsCA"

CA_NAME = 'Self-Signed-CA'
CERT_NAME = 'Server-Cert'

class DynamicCert(DSLdapObject):
    """
    Represents a single DynamicCert LDAP entry.
    """

    _must_attributes = [DYCATTR_CN]

    def __init__(self, instance, dn: Optional[str] = None):
        """
        Initialise a DynamicCert object.

        :param instance: DirSrv instance
        :param dn: Entry distinguished name (DN)
        """
        super(DynamicCert, self).__init__(instance, dn)
        self._rdn_attribute = DYCATTR_CN
        self._create_objectclasses = ["top", "extensibleObject"]
        self._protected = False
        self._basedn = DYNCERT_SUFFIX

    def _normalise_timestamp(self, raw: str):
        """
        Convert DynamicCert timestamp to NSS like format.

        :param raw: Raw DynamicCert timestamp (e.g. 20260109181934Z)
        :return: Formatted timestamp as "YYYY-MM-DD HH:MM:SS", or original string
        """
        try:
            year   = int(raw[0:4])
            month  = int(raw[4:6]) + 1  # PRExplodedTime.tm_month is 0–11
            day    = int(raw[6:8])
            hour   = int(raw[8:10])
            minute = int(raw[10:12])
            second = int(raw[12:14])
            return datetime.datetime(year, month, day, hour, minute, second).strftime("%Y-%m-%d %H:%M:%S")
        except Exception:
            return raw

    def del_cert(self):
        """
        Delete this DynamicCert entry from LDAP.

        :raises ValueError: If the DynamicCert object does not have a DN
        :raises ldap.LDAPError: If an LDAP operation fails (other than NO_SUCH_OBJECT)
        """
        if not self._dn:
            raise ValueError("Cannot delete DynamicCert without a DN")

        try:
            self.delete()
        except ldap.NO_SUCH_OBJECT:
            log.warning(f"DynamicCert already deleted: {self._dn}")
        except ldap.LDAPError as e:
            log.error(f"Failed to delete DynamicCert: {self._dn}: {e}")
            raise

    def edit_trust(self, trust_flags: str):
        """
        Edit certificate trust flags.

        :param trust_flags: Comma separated trust flags string (SSL,Email,ObjectSigning)
        :raises ValueError: If trust flags are invalid or empty
        """
        if not trust_flags:
            raise ValueError("Trust flags cannot be empty")

        trust_fields = trust_flags.strip().split(",")

        if len(trust_fields) != 3:
            raise ValueError("Trust flags must have 3 comma separated fields")

        # Allowed field values (NSS)
        valid_flags = set("pPcCTu")
        for field in trust_fields:
            if field and any(flag not in valid_flags for flag in field):
                raise ValueError(f"Invalid characters in trust flags: '{trust_flags}'")

        try:
            self.replace(DYCATTR_TRUST, trust_flags)
        except ldap.LDAPError as e:
            log.error(f"Failed to update trust flags for {self._dn}: {e}")
            raise

class DynamicCerts(DSLdapObjects):
    """
    Collection of DynamicCert entries under cn=dynamiccertificates.
    """

    def __init__(self, instance):
        """
        Initialise the DynamicCerts collection.

        :param instance: DirSrv instance
        """
        super(DynamicCerts, self).__init__(instance=instance)
        self._objectclasses = ["extensibleObject"]
        self._filterattrs = [DYCATTR_CN, DYCATTR_SUBJECT, DYCATTR_ISSUER]
        self._childobject = DynamicCert
        self._basedn = DYNCERT_SUFFIX

    def add_cert(self,
                 cert_file: str,
                 nickname: str,
                 pkcs12_password: Optional[str] = None,
                 ca: bool = False,
                 force: bool = False,
    ):
        """
        Add or update a certificate (PEM, DER, or PKCS#12).

        :param cert_file: Path to certificate file (PEM, DER, or PKCS#12)
        :param nickname: Certificate nickname
        :param pkcs12_password: Password for PKCS#12, if any
        :param ca: Whether this certificate is a CA certificate
        :param force: Force the addition of a certificate that cannot be verified
        """
        if not nickname:
            raise ValueError("Certificate CN cannot be empty")

        if not os.path.isfile(cert_file):
            raise ValueError(f"Certificate file does not exist: {cert_file}")

        if pkcs12_password and not isinstance(pkcs12_password, str):
            raise TypeError("PKCS#12 password must be a string")

        with open(cert_file, "rb") as f:
            cert_bytes = f.read()

        der_cert = None
        der_privkey = None

        if cert_file.lower().endswith((".p12", ".pfx")):
            try:
                privkey, cert, _ = pkcs12.load_key_and_certificates(
                    cert_bytes, pkcs12_password.encode() if pkcs12_password else None,
                    backend=default_backend()
                )
            except Exception as e:
                raise ValueError(f"Failed to load PKCS#12 file: {cert_file}: {e}")

            if cert is None:
                raise ValueError("PKCS#12 file contains no certificate")

            der_cert = cert.public_bytes(serialization.Encoding.DER)
            if privkey is not None:
                der_privkey = privkey.private_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption()
                )
        else:
            try:
                der_cert = pem_to_der(cert_bytes) if is_pem_cert(cert_bytes) else cert_bytes
            except Exception as e:
                raise ValueError(f"Failed to parse certificate '{cert_file}': {e}")

        if ca and not cert_is_ca(cert_file):
            raise ValueError(f"Certificate ({nickname}) is not a CA certificate")

        attrs = {
            "cn": [nickname.encode()],
            "objectClass": [b"top", b"extensibleObject"],
            DYCATTR_CERTDER: [ der_cert, ]
        }
        if der_privkey:
            attrs[DYCATTR_PKEYDER] = [der_privkey]

        if ca:
            attrs[DYCATTR_TRUST] = [b"CT,,"]

        if force:
            attrs[DYCATTR_FORCE] = [b"TRUE"]

        if not der_cert:
            raise ValueError(f"Failed to extract DER bytes from {cert_file}")

        # Escape the CN to handle special chars in the nickname
        escaped_cn = ldap.dn.escape_dn_chars(nickname)
        dn = ensure_str(f"cn={escaped_cn},{self._basedn}")
        # Raw CN used for lookup
        cert_obj = self.get_cert_obj(nickname)
        if not cert_obj:
            cert_obj = DynamicCert(self._instance, dn)
            cert_obj.create(properties=attrs)
        else:
            log.info(f"Updating existing certificate; {nickname}")
            attrs_list = [(attr, vals) for attr, vals in attrs.items()]
            cert_obj.replace_many(*attrs_list)

    def add_ca_cert(self,
                    cert_file: str,
                    nickname: str,
                    pkcs12_password: Optional[str] = None,
                    force: bool = False
        ):
        """
        Add a CA certificate from a PEM bundle or single PEM/DER file.

        :param cert_file: Path to the certificate file (PEM or DER)
        :param nickname: Certificate nickname
        :param pkcs12_password: Password for PKCS#12, if any
        :param force: Force the addition of a certificate that cannot be verified
        :raises ValueError: If file is invalid or file does not exist
        """
        if not os.path.exists(cert_file):
            raise ValueError(f"Certificate file does not exist: {cert_file}")

        # Normalise nickname(s)
        if isinstance(nickname, str):
            nicknames = [nickname]
        elif isinstance(nickname, list):
            nicknames = nickname
        else:
            raise TypeError(f"nickname must be str or list[str], got {type(nickname)}")

        # PEM (may be bundle)
        if cert_file.lower().endswith(".pem"):
            with open(cert_file, "r") as f:
                pem_data = f.read()

            pem_certs = re.findall(
                r"-----BEGIN CERTIFICATE-----.*?-----END CERTIFICATE-----",
                pem_data,
                re.DOTALL
            )

            if not pem_certs:
                raise ValueError("No certificates found in PEM file")

            temp_files = []
            try:
                for idx, cert in enumerate(pem_certs):
                    with tempfile.NamedTemporaryFile(delete=False, suffix=".pem", mode="w") as tmp:
                        tmp.write(cert.strip() + "\n")
                        tmp_cert_path = tmp.name
                    temp_files.append(tmp_cert_path)

                    # Determine nickname, could be a list, or not enough
                    if idx < len(nicknames):
                        ca_nick = nicknames[idx]
                    else:
                        ca_nick = f"{nicknames[-1]}{idx}"

                    # Check we dont trample over installer certs
                    if ca_nick.lower() in (CERT_NAME.lower(), CA_NAME.lower()):
                        raise ValueError(f"You may not import a CA with the nickname {CERT_NAME} or {CA_NAME}")

                    if not cert_is_ca(tmp_cert_path):
                        raise ValueError(f"Certificate ({ca_nick}) is not a CA certificate")

                    try:
                        # Handle existing LDAP object
                        if self.get_cert_details(ca_nick):
                            if not force:
                                raise ValueError(
                                    f"Certificate already exists with the same name ({ca_nick})"
                                )
                            else:
                                log.info(f"Overwriting existing CA cert {ca_nick}")
                                self.del_cert(ca_nick)
                    except ValueError:
                        pass

                    self.add_cert(
                        tmp_cert_path,
                        ca_nick,
                        pkcs12_password=pkcs12_password,
                        ca=True,
                        force=force
                    )

            finally:
                for tmp_file in temp_files:
                    try:
                        os.remove(tmp_file)
                    except OSError as e:
                        log.debug(f"Failed to remove tmp cert file: {tmp_file}: {e}")
        else:
            # Single binary cert
            if len(nicknames) != 1:
                raise ValueError("Single cert requires exactly one nickname")

            ca_nick = nicknames[0]
            try:
                if self.get_cert_details(ca_nick):
                    if force:
                        self.del_cert(ca_nick)
                    else:
                        raise ValueError(f"Certificate already exists: {ca_nick}")
            except ValueError:
                pass

            self.add_cert(cert_file, ca_nick, pkcs12_password=pkcs12_password, ca=True, force=force)

    def del_cert(self, nickname: str):
        """
        Delete a certificate.

        :param nickname: The certificate nickname to delete.
        :raises ValueError: If the nickname is empty or the entry cannot be found.
        """
        if not nickname:
            raise ValueError("Certificate nickname cannot be empty")

        cert_obj = self.get_cert_obj(nickname)
        if not cert_obj:
            raise ValueError(f"Certificate entry not found: {nickname}")

        cert_obj.del_cert()

    def list_certs(self):
        """
        List all server certificates.

        :return: A list of certificate dictionaries for each certificate
        """
        cert_objects = self.list()
        certs = []
        for cert in cert_objects:
            if cert._dn == self._basedn:
                continue
            der_cert = cert.get_attr_vals_bytes(DYCATTR_CERTDER)[0]
            if der_cert:
                if not cert_is_ca(der_cert):
                    certs.append({
                        "cn": cert.get_attr_vals_utf8(DYCATTR_CN)[0],
                        "subject": cert.get_attr_vals_utf8(DYCATTR_SUBJECT)[0],
                        "issuer": cert.get_attr_vals_utf8(DYCATTR_ISSUER)[0],
                        "expires": cert._normalise_timestamp(cert.get_attr_vals_utf8(DYCATTR_NOTAFTER)[0]),
                        "trust_flags": cert.get_attr_vals_utf8(DYCATTR_TRUST)[0],
                    })
        return certs

    def list_ca_certs(self):
        """
        List all ca certificates.

        :return: A list of certificate dictionaries for each certificate
        """
        cert_objects = self.list()
        certs = []
        for cert in cert_objects:
            if cert._dn == self._basedn:
                continue
            der_cert = cert.get_attr_vals_bytes(DYCATTR_CERTDER)[0]
            if der_cert:
                if cert_is_ca(der_cert):
                    certs.append({
                        "cn": cert.get_attr_vals_utf8(DYCATTR_CN)[0],
                        "subject": cert.get_attr_vals_utf8(DYCATTR_SUBJECT)[0],
                        "issuer": cert.get_attr_vals_utf8(DYCATTR_ISSUER)[0],
                        "expires": cert._normalise_timestamp(cert.get_attr_vals_utf8(DYCATTR_NOTAFTER)[0]),
                        "trust_flags": cert.get_attr_vals_utf8(DYCATTR_TRUST)[0],
                    })
        return certs

    def get_cert_obj(self, nickname: str):
        """
        Retrieve a certificate object.

        :param cn: Certificate nickname
        :raises ValueError: If the cn is empty
        :return: DynamicCert object if found, else None
        """
        if not nickname:
            raise ValueError("Certificate CN cannot be empty")
        try:
            cert = self.get(nickname)
        except ldap.NO_SUCH_OBJECT:
            return None

        return cert

    def get_cert_details(self, nickname: str):
        """
        Get a certificates details.

        :param nickname: Certificate nickname
        :raises ValueError: If the nickname is empty
        :return: DynamicCert object if found, else None
        """
        if not nickname:
            raise ValueError("Certificate CN cannot be empty")
        try:
            cert = self.get(nickname)
        except ldap.NO_SUCH_OBJECT:
            return None

        return {
            "cn": cert.get_attr_vals_utf8(DYCATTR_CN)[0],
            "subject": cert.get_attr_vals_utf8(DYCATTR_SUBJECT)[0],
            "issuer": cert.get_attr_vals_utf8(DYCATTR_ISSUER)[0],
            "expires": cert._normalise_timestamp(cert.get_attr_vals_utf8(DYCATTR_NOTAFTER)[0]),
            "trust_flags": cert.get_attr_vals_utf8(DYCATTR_TRUST)[0],
        }

    def edit_cert_trust(self, nickname: str, trust_flags: str):
        """
        Edit trust flags on an existing certificate.

        :param nickname: Certificate nickname
        :param trust_flags: 3 field NSS trust string
        """
        cert_obj = self.get_cert_obj(nickname)
        if not cert_obj:
            raise ValueError(f"Certificate {nickname} does not exist")

        try:
            cert_obj.edit_trust(trust_flags=trust_flags)
        except ValueError as ve:
            log.error(f"Invalid input for certificate '{nickname}': {ve}")
            raise
        except ldap.LDAPError as le:
            log.error(f"LDAP error while updating certificate '{nickname}': {le}")
            raise
        except Exception as e:
            log.error(f"Unexpected error while updating certificate '{nickname}': {e}")
            raise



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# Copyright (C) 2019 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import argparse
import ast
import logging
import sys
import json
import ldap
from ldap.dn import is_dn
from getpass import getpass
from lib389 import DirSrv
from lib389.utils import assert_c
from lib389.dseutils import get_ldapurl_from_serverid
from lib389.properties import SER_ROOT_PW, SER_ROOT_DN


def _get_arg(args, msg=None, hidden=False, confirm=False):
    if args is not None and len(args) > 0:
        if type(args) is list:
            return args[0]
        else:
            return args
    else:
        if hidden:
            if confirm:
                while True:
                    x = getpass("%s : " % msg)
                    y = getpass("CONFIRM - %s : " % msg)
                    if x != y:
                        print("Passwords do not match, try again.")
                    else:
                        return y
            else:
                return getpass("%s : " % msg)
        else:
            return input("%s : " % msg)


def _get_dn_arg(args, msg=None):
    dn_arg = _get_arg(args, msg)
    if not is_dn(dn_arg):
        raise ValueError(f"{dn_arg} is not a valid DN")
    return dn_arg


def _get_args(args, kws):
    kwargs = {}
    while len(kws) > 0:
        kw, msg, priv = kws.pop(0)
        kw = kw.lower()
        if args is not None and len(args) > 0:
            kwargs[kw] = args.pop(0)
        else:
            if priv:
                kwargs[kw] = getpass("%s : " % msg)
            else:
                kwargs[kw] = input("%s : " % msg)
    return kwargs


# This is really similar to get_args, but generates from an array
def _get_attributes(args, attrs):
    kwargs = {}
    for attr in attrs:
        # Python can't represent a -, so it replaces it to _
        # in many places, so we have to normalise this.
        attr_normal = attr.replace('-', '_')
        if args is not None and hasattr(args, attr_normal) and getattr(args, attr_normal) is not None:
            attr = attr.lower()
            kwargs[attr] = getattr(args, attr_normal)
        else:
            if attr.lower() == 'userpassword':
                kwargs[attr] = getpass("Enter value for %s : " % attr)
            else:
                attr_normal = attr.lower()
                kwargs[attr_normal] = input("Enter value for %s : " % attr)

    return kwargs


def _warn(data, msg=None):
    if msg is not None:
        print("%s :" % msg)
    if 'Yes I am sure' != input("Type 'Yes I am sure' to continue: "):
        raise Exception("Not sure I want")
    return data


def connect_instance(dsrc_inst, verbose, args):
    dsargs = dsrc_inst['args']
    if '//' not in dsargs['ldapurl']:
        # Connecting to the local instance
        dsargs['server-id'] = dsargs['ldapurl']
        # We have an instance name - generate url from dse.ldif
        ldapurl, certdir = get_ldapurl_from_serverid(dsargs['ldapurl'])
        if ldapurl is not None:
            dsargs['ldapurl'] = ldapurl
            if 'ldapi://' in ldapurl:
                dsargs['ldapi_enabled'] = 'on'
                dsargs['ldapi_socket'] = ldapurl.replace('ldapi://', '')
                dsargs['ldapi_autobind'] = 'on'
            elif 'ldaps://' in ldapurl:
                dsrc_inst['tls_cert'] = certdir
        else:
            # The instance name does not match any instances
            raise ValueError("Could not find configuration for instance: " + dsargs['ldapurl'])

    ds = DirSrv(verbose=verbose)
    # We do an empty allocate here to determine if we can autobind ... (really
    # we should actually be inspect the URL ...)
    ds.allocate(dsargs)

    if args.pwdfile is not None or args.bindpw is not None or args.prompt is True:
        if args.pwdfile is not None:
            # Read password from file
            try:
                with open(args.pwdfile, "r") as f:
                    dsargs[SER_ROOT_PW] = f.readline().rstrip()
            except EnvironmentError as e:
                raise ValueError("Failed to open password file: " + str(e))
        elif args.bindpw is not None:
            # Password provided
            # This shouldn't be needed? dsrc already inherits the args ...
            dsargs[SER_ROOT_PW] = args.bindpw
        else:
            # No password or we chose to prompt
            dsargs[SER_ROOT_PW] = getpass("Enter password for {} on {}: ".format(dsrc_inst['binddn'], dsrc_inst['uri']))
    elif not ds.can_autobind():
        # No LDAPI, prompt for password, and bind DN if necessary
        if dsrc_inst['binddn'] is None:
            dn = ""
            while dn == "":
                dn = input("Enter Bind DN: ")
            dsrc_inst['binddn'] = dn
        dsargs[SER_ROOT_PW] = getpass("Enter password for {} on {}: ".format(dsrc_inst['binddn'], dsrc_inst['uri']))

    # Allocate is an awful interface that we should stop using, but for now
    # just directly map the dsrc_inst args in (remember, dsrc_inst DOES
    # overlay cli args into the map ...)
    dsargs[SER_ROOT_DN] = dsrc_inst['binddn']

    ds = DirSrv(verbose=verbose)
    ds.allocate(dsargs)
    ds.open(saslmethod=dsrc_inst['saslmech'],
            certdir=dsrc_inst['tls_cacertdir'],
            reqcert=dsrc_inst['tls_reqcert'],
            usercert=dsrc_inst['tls_cert'],
            userkey=dsrc_inst['tls_key'],
            starttls=dsrc_inst['starttls'], connOnly=True)
    if ds.serverid is not None and ds.serverid.startswith("slapd-"):
        ds.serverid = ds.serverid.replace("slapd-", "", 1)
    return ds


def disconnect_instance(inst):
    if inst is not None:
        inst.close()


def populate_attr_arguments(parser, attributes):
    for attr in attributes:
        parser.add_argument('--%s' % attr, nargs='?', help="Value of %s" % attr)


def _format_status(log, mtype, json=False):
    if json:
        print(mtype.get_status_json())
    else:
        status_dict = mtype.get_status()
        log.info('dn: ' + mtype._dn)
        for k, v in list(status_dict.items()):
            # For each value in the multivalue attr
            for vi in v:
                log.info('{}: {}'.format(k, vi))


def _format_status_with_option(log, mtype, json=False, option=False):
    if json:
        print(mtype.get_status_json(option))
    else:
        status_dict = mtype.get_status(option)
        log.info('dn: ' + mtype._dn)
        for k, v in list(status_dict.items()):
            # For each value in the multivalue attr
            for vi in v:
                log.info('{}: {}'.format(k, vi))

def _generic_list(inst, basedn, log, manager_class, args=None):
    mc = manager_class(inst, basedn)
    ol = mc.list()
    if len(ol) == 0:
        if args and args.json:
            log.info(json.dumps({"type": "list", "items": []}, indent=4))
        else:
            log.info("No objects to display")
    elif len(ol) > 0:
        # We might sort this in the future
        if args and args.json:
            json_result = {"type": "list", "items": []}
        for o in ol:
            o_str = o.__unicode__()
            if args and args.json:
                json_result['items'].append(o_str)
            else:
                log.info(o_str)
        if args and args.json:
            log.info(json.dumps(json_result, indent=4))


# Display these entries better!
def _generic_get(inst, basedn, log, manager_class, selector, args=None):
    mc = manager_class(inst, basedn)
    if args and args.json:
        o = mc.get(selector, json=True)
        log.info(o)
    else:
        o = mc.get(selector)
        o_str = o.display()
        log.info(o_str)


def _generic_get_entry(inst, basedn, log, manager_class, args=None):
    mc = manager_class(inst, basedn)
    if args and args.json:
        log.info(mc.get_all_attrs_json())
    else:
        log.info(mc.display())


def _generic_get_attr(inst, basedn, log, manager_class, args=None):
    mc = manager_class(inst, basedn)
    vals = {}
    for attr in args.attrs:
        if args and args.json:
            vals[attr] = mc.get_attr_vals_utf8(attr)
        else:
            log.info(mc.display_attr(attr).rstrip())
    if args.json:
        log.info(json.dumps({"type": "entry", "dn": mc._dn, "attrs": vals}, indent=4))


def _generic_get_dn(inst, basedn, log, manager_class, dn, args=None):
    mc = manager_class(inst, basedn)
    if args is not None and args.json:
        o = mc.get(dn=dn, json=True)
        log.info(o)
    else:
        o = mc.get(dn=dn)
        log.info(o.display())


def _generic_create(inst, basedn, log, manager_class, kwargs, args=None):
    mc = manager_class(inst, basedn)
    o = mc.create(properties=kwargs)
    o_str = o.__unicode__()
    log.info('Successfully created %s' % o_str)


def _generic_delete(inst, basedn, log, object_class, dn, args=None):
    # Load the oc direct
    o = object_class(inst, dn)
    o.delete()
    log.info('Successfully deleted %s' % dn)


# Attr functions expect attribute values to be "attr=value"

def _generic_replace_attr(inst, basedn, log, manager_class, args=None):
    mc = manager_class(inst, basedn)
    if args and args.attr:
        for myattr in args.attr:
            if "=" in myattr:
                [attr, val] = myattr.split("=", 1)
                mc.replace(attr, val)
                log.info("Successfully replaced \"{}\"".format(attr))
            else:
                raise ValueError("You must specify a value to replace the attribute ({})".format(myattr))
    else:
        # Missing value
        raise ValueError("Missing attribute to replace")


def _generic_add_attr(inst, basedn, log, manager_class, args=None):
    mc = manager_class(inst, basedn)
    if args and args.attr:
        for myattr in args.attr:
            if "=" in myattr:
                [attr, val] = myattr.split("=", 1)
                mc.add(attr, val)
                log.info("Successfully added \"{}\"".format(attr))
            else:
                raise ValueError("You must specify a value to add for the attribute ({})".format(myattr))
    else:
        # Missing value
        raise ValueError("Missing attribute to add")


def _generic_del_attr(inst, basedn, log, manager_class, args=None):
    mc = manager_class(inst, basedn)
    if args and args.attr:
        for myattr in args.attr:
            if "=" in myattr:
                # we have a specific value
                [attr, val] = myattr.split("=", 1)
                mc.remove(attr, val)
            else:
                # remove all
                mc.remove_all(myattr)
                attr = myattr  # for logging
            log.info("Successfully removed \"{}\"".format(attr))
    else:
        # Missing value
        raise ValueError("Missing attribute to delete")


def _generic_modify_change_to_mod(change):
    values = change.split(":")
    if len(values) <= 2:
        raise ValueError("Not enough arguments in '%s'. action:attribute:value" % change)
    elif len(values) >= 4:
        raise ValueError("Too many arguments in '%s'. %s:attribute:value expected" % (change, values[0]))
    elif len(values[1]) == 0:
        raise ValueError("Invalid empty attribute name in '%s'." % change)
    # Return a tuple
    if values[0] == 'add':
        if len(values[2]) == 0:
            raise ValueError("Invalid empty value in '%s'." % change)
        return (ldap.MOD_ADD, values[1], values[2])
    elif values[0] == 'delete':
        if len(values[2]) == 0:
            return (ldap.MOD_DELETE, values[1])
        return (ldap.MOD_DELETE, values[1], values[2])
    elif values[0] == 'replace':
        if len(values[2]) == 0:
            raise ValueError("Invalid empty value in '%s'." % change)
        return (ldap.MOD_REPLACE, values[1], values[2])
    else:
        raise ValueError("Unknown action '%s'. Expected add, delete, replace" % change)


def _generic_modify_inner(log, o, changes):
    # Now parse the series of arguments.
    # Turn them into mod lists. See apply_mods.
    mods = [_generic_modify_change_to_mod(x) for x in changes]
    log.debug("Requested mods: %s" % mods)
    # Now push them to dsldapobject to modify
    o.apply_mods(mods)
    log.info('Successfully modified %s' % o.dn)


def _generic_modify(inst, basedn, log, manager_class, selector, args=None):
    if not args or not args.changes:
        raise ValueError("Missing modify actions to perform.")
    # Here, we should have already selected the type etc. mc should be a
    # type of DSLdapObjects (plural)
    mc = manager_class(inst, basedn)
    # Get the object singular by selector
    try:
        o = mc.get(selector)
    except ldap.NO_SUCH_OBJECT:
        raise ValueError(f'The entry, or base DN ({mc._basedn}) does not exist')
    _generic_modify_inner(log, o, args.changes)


def _generic_modify_dn(inst, basedn, log, manager_class, dn, args=None):
    if not args or not args.changes:
        raise ValueError("Missing modify actions to perform.")
    # Here, we should have already selected the type etc. mc should be a
    # type of DSLdapObjects (plural)
    mc = manager_class(inst, basedn)
    # Get the object singular by dn
    o = mc.get(dn=dn)
    _generic_modify_inner(log, o, args.changes)


# We need parent parser to be able to add -v and -j to all subparsers
# because we use parent_arguments in CustomHelpFormatter
parent_arguments = []
parent_argparser = argparse.ArgumentParser(add_help=False)
parent_arguments.append(parent_argparser.add_argument('-v', '--verbose',
        help="Display verbose operation tracing during command execution",
        action='store_true', default=False
    ))
parent_arguments.append(parent_argparser.add_argument('-j', '--json',
        help="Return result in JSON object",
        default=False, action='store_true'
    ))


class CustomHelpFormatter(argparse.HelpFormatter):
    """Custom help formatter to add [-v] [-j] to the usage line and add these options'
    description to the full help output
    """
    def add_arguments(self, actions):
        if len(actions) > 0:
            # Check if this is the main options section by looking for the help action
            is_main_section = any(
                isinstance(action, argparse._HelpAction)
                for action in actions
            )

            # Only add parent arguments to the main options section
            if is_main_section:
                actions = parent_arguments + actions

        super(CustomHelpFormatter, self).add_arguments(actions)

    def _format_usage(self, usage, actions, groups, prefix):
        usage = super(CustomHelpFormatter, self)._format_usage(usage, actions, groups, prefix)

        if sys.version_info < (3, 13):
            # Use _format_actions_usage() for Python 3.12 and earlier
            formatted_options = self._format_actions_usage(parent_arguments, [])
        else:
            # Use _get_actions_usage_parts() for Python 3.13 and later
            action_parts = self._get_actions_usage_parts(parent_arguments, [])
            if isinstance(action_parts, tuple):
                # Python 3.14.3+ and 3.15+ return a tuple (list of actions, count of actions)
                formatted_options = ' '.join(action_parts[0])
            else:
                # Earlier versions return a list of actions
                formatted_options = ' '.join(action_parts)

        # If formatted_options already in usage - remove them
        if formatted_options in usage:
            usage = usage.replace(f' {formatted_options}', '')
        usage = usage.split(' ')
        usage.insert(2, formatted_options)
        usage = ' '.join(usage)

        return usage


class LogCapture(logging.Handler):
    """
    This useful class is for intercepting logs, and then making assertions about
    the outputs provided. Used by the cli unit tests.
    """

    def __init__(self):
        """
        Create a log instance and primes the output capture.
        """
        super(LogCapture, self).__init__()
        self.outputs = []
        self.raw_outputs = []
        self.log = logging.getLogger("LogCapture")
        self.log.addHandler(self)
        self.log.setLevel(logging.INFO)

    def emit(self, record):
        self.outputs.append(record)
        self.raw_outputs.append(str(record.msg))

    def contains(self, query):
        """
        Assert that the query string listed is in some logged Record.
        """
        result = False
        for rec in self.outputs:
            if query in str(rec):
                result = True
        return result

    def print(self):
        for rec in self.outputs:
            print(str(rec))

    def get_raw_outputs(self):
        return self.raw_outputs

    def flush(self):
        self.outputs = []
        self.raw_outputs = []


class FakeArgs(object):
    def __init__(self):
        pass

    def __len__(self):
        return len(self.__dict__.keys())


class StdErrFilter(logging.Filter):
    def filter(self, rec):
        return rec.levelno in (logging.ERROR,)


class StdOutFilter(logging.Filter):
    def filter(self, rec):
        return rec.levelno not in (logging.ERROR,)


def setup_script_logger(name, verbose=False):
    """Reset the python logging system for STDOUT, and attach a new
    console logger with cli expected formatting.

    :param name: Name of the logger
    :type name: str
    :param verbose: Enable verbose format of messages
    :type verbose: bool
    :return: logging.logger
    """
    root = logging.getLogger()
    log = logging.getLogger(name)
    log_handler = logging.StreamHandler(sys.stdout)

    if verbose:
        log.setLevel(logging.DEBUG)
        log_format = '%(levelname)s: %(message)s'
    else:
        log.setLevel(logging.INFO)
        log_format = '%(message)s'

    log_handler.setFormatter(logging.Formatter(log_format))
    log_handler.addFilter(StdOutFilter())
    root.addHandler(log_handler)

    log_handler_err = logging.StreamHandler(sys.stderr)
    log_handler_err.setLevel(logging.ERROR)
    log_handler_err.addFilter(StdErrFilter())
    log_handler_err.setFormatter(logging.Formatter(log_format))
    log.addHandler(log_handler_err)

    return log


def format_error_to_dict(exception):
    """python-ldap str(exception) processing is not consistent.
    This function makes sure that the result is dict

    :param exception: Exception you need to print
    :type exception: ldap.LDAPError
    :returns: dict
    """

    # The issue was raise on python-ldap repo: https://github.com/python-ldap/python-ldap/issues/304
    # We should fix the code here after the issue is fixed
    errmsg = str(exception)
    try:
        # The function literal_eval parses the string and returns only if it's
        # a literal. Also, the code is never executed. So there is no reason
        # for a security risk.
        msg = ast.literal_eval(errmsg)
    except Exception:
        msg = {'desc': errmsg}

    return msg


def format_pretty_error(msg_dict):
    """
    Take a raw exception dict and make a pretty message for the user
    """
    msg = f"{msg_dict['desc']}"
    if 'result' in msg_dict:
        msg += f"({msg_dict['result']})"
    if 'info' in msg_dict:
        msg += f" - {msg_dict['info']}"

    return {'desc': msg}



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# Copyright (C) 2022 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
import ldap
from lib389.plugins import (PassThroughAuthenticationPlugin)
from lib389.cli_base import CustomHelpFormatter
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add, generic_show, generic_enable, generic_disable, generic_status


def _get_url_next_num(url_attrs):
    existing_nums = list(map(lambda url: int(url.split('nsslapd-pluginarg')[1]),
                             [i for i, _ in url_attrs.items()]))
    if len(existing_nums) > 0:
        existing_nums.sort()
        full_num_list = list(range(existing_nums[-1]+2))
        if not full_num_list:
            next_num_list = ["0"]
        else:
            next_num_list = list(filter(lambda x: x not in existing_nums, full_num_list))
    else:
        next_num_list = ["0"]

    return next_num_list[0]

def _get_url_attr(url_attrs,  this_url):
    for attr, val in url_attrs.items():
        if val.lower() == this_url.lower():
            return attr

    return "nsslapd-pluginarg0"

def _validate_url(url):
    failed = False
    if len(url.split(" ")) == 2:
        link = url.split(" ")[0]
        params = url.split(" ")[1]
    else:
        link = url
        params = ""

    if (":" not in link) or ("//" not in link) or ("/" not in link) or (params and "," not in params):
        failed = False

    if not ldap.dn.is_dn(link.split("/")[-1]):
        raise ValueError("Subtree is an invalid DN")

    if params and len(params.split(",")) != 6 and not all(map(str.isdigit, params.split(","))):
        failed = False

    if failed:
        raise ValueError("URL should be in one of the next formats (all parameters after a space should be digits): "
                         "'ldap|ldaps://authDS/subtree maxconns,maxops,timeout,ldver,connlifetime,startTLS' or "
                         "'ldap|ldaps://authDS/subtree'")
    return url


def pta_list(inst, basedn, log, args):
    log = log.getChild('pta_list')
    plugin = PassThroughAuthenticationPlugin(inst)
    urls = plugin.get_urls()
    if args.json:
        log.info(json.dumps({"type": "list",
                             "items": [{"id": id, "url": value} for id, value in urls.items()]},
                            indent=4))
    else:
        if len(urls) > 0:
            for _, value in urls.items():
                log.info(value)
        else:
            log.info("No Pass Through Auth URLs were found")


def pta_add(inst, basedn, log, args):
    log = log.getChild('pta_add')
    new_url_l = _validate_url(args.URL.lower())
    plugin = PassThroughAuthenticationPlugin(inst)
    url_attrs = plugin.get_urls()
    urls = list(map(lambda url: url.lower(),
                    [i for _, i in url_attrs.items()]))
    next_num = _get_url_next_num(url_attrs)
    if new_url_l in urls:
        raise ldap.ALREADY_EXISTS("Entry %s already exists" % args.URL)
    plugin.add("nsslapd-pluginarg%s" % next_num, args.URL)
    log.info("Successfully added URL")


def pta_edit(inst, basedn, log, args):
    log = log.getChild('pta_edit')
    plugin = PassThroughAuthenticationPlugin(inst)
    url_attrs = plugin.get_urls()
    urls = list(map(lambda url: url.lower(),
                    [i for _, i in url_attrs.items()]))
    _validate_url(args.NEW_URL.lower())
    old_url_l = args.OLD_URL.lower()
    if old_url_l not in urls:
        raise ValueError("URL %s doesn't exist." % args.OLD_URL)
    else:
        for attr, value in url_attrs.items():
            if value.lower() == old_url_l:
                plugin.remove(attr, old_url_l)
                break
    plugin.add("%s" % _get_url_attr(url_attrs, old_url_l), args.NEW_URL)


def pta_del(inst, basedn, log, args):
    log = log.getChild('pta_del')
    plugin = PassThroughAuthenticationPlugin(inst)
    url_attrs = plugin.get_urls()
    urls = list(map(lambda url: url.lower(),
                    [i for _, i in url_attrs.items()]))
    old_url_l = args.URL.lower()
    if old_url_l not in urls:
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % args.URL)

    plugin.remove_all("%s" % _get_url_attr(url_attrs, old_url_l))
    log.info("Successfully deleted %s", args.URL)


def create_parser(subparsers):
    passthroughauth_parser = subparsers.add_parser('ldap-pass-through-auth',
                                                   help='Manage and configure LDAP Pass-Through Authentication Plugin')
    subcommands = passthroughauth_parser.add_subparsers(help='action')

    add_generic_plugin_parsers(subcommands, PassThroughAuthenticationPlugin)

    list_urls = subcommands.add_parser('list', help='Lists LDAP URLs', formatter_class=CustomHelpFormatter)
    list_urls.set_defaults(func=pta_list)

    # url = subcommands.add_parser('url', help='Manage PTA LDAP URL configurations', formatter_class=CustomHelpFormatter)
    # subcommands_url = url.add_subparsers(help='action')

    add_url = subcommands.add_parser('add', help='Add an LDAP url to the config entry', formatter_class=CustomHelpFormatter)
    add_url.add_argument('URL',
                         help='The full LDAP URL in format '
                              '"ldap|ldaps://authDS/subtree maxconns,maxops,timeout,ldver,connlifetime,startTLS". '
                              'If one optional parameter is specified the rest should be specified too')
    add_url.set_defaults(func=pta_add)

    edit_url = subcommands.add_parser('modify', help='Edit the LDAP pass through config entry', formatter_class=CustomHelpFormatter)
    edit_url.add_argument('OLD_URL', help='The full LDAP URL you get from the "list" command')
    edit_url.add_argument('NEW_URL',
                          help='Sets the full LDAP URL in format '
                               '"ldap|ldaps://authDS/subtree maxconns,maxops,timeout,ldver,connlifetime,startTLS". '
                               'If one optional parameter is specified the rest should be specified too.')
    edit_url.set_defaults(func=pta_edit)

    delete_url = subcommands.add_parser('delete', help='Delete a URL from the config entry', formatter_class=CustomHelpFormatter)
    delete_url.add_argument('URL', help='The full LDAP URL you get from the "list" command')
    delete_url.set_defaults(func=pta_del)




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016-2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest

from lib389.tests.cli import topology as default_topology
from lib389.cli_base import LogCapture, FakeArgs
from lib389.plugins import USNPlugin
from lib389.cli_conf.plugins import usn as usn_cli


@pytest.fixture(scope="module")
def topology(request):
    topology = default_topology(request)

    plugin = USNPlugin(topology.standalone)
    if not plugin.exists():
        plugin.create()

    # we need to restart the server after enabling the plugin
    plugin.enable()
    topology.standalone.restart()
    topology.logcap.flush()

    return topology


def test_enable_global_mode(topology):
    args = FakeArgs()

    usn_cli.enable_global_mode(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("USN global mode enabled")
    topology.logcap.flush()

    usn_cli.display_usn_mode(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("USN global mode is enabled")
    topology.logcap.flush()

def test_disable_global_mode(topology):
    args = FakeArgs()

    usn_cli.disable_global_mode(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("USN global mode disabled")
    topology.logcap.flush()

    usn_cli.display_usn_mode(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("USN global mode is disabled")
    topology.logcap.flush()
